import warnings
from typing import Any, Dict, List, Optional, Set, Union

from sqllineage.config import SQLLineageConfig
from sqllineage.exceptions import SQLLineageException
from sqllineage.utils.helpers import escape_identifier_name


class Schema:
    """
    Data Class for Schema
    """

    unknown = "<default>"

    def __init__(self, name: Optional[str] = None):
        """
        :param name: schema name
        """
        if name:
            self.raw_name = escape_identifier_name(name)
        elif SQLLineageConfig.DEFAULT_SCHEMA:
            self.raw_name = escape_identifier_name(SQLLineageConfig.DEFAULT_SCHEMA)
        else:
            self.raw_name = escape_identifier_name(Schema.unknown)

    def __str__(self):
        return self.raw_name

    def __repr__(self):
        return "Schema: " + str(self)

    def __eq__(self, other):
        return isinstance(other, Schema) and str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def __bool__(self):
        return str(self) != self.unknown


class Table:
    """
    Data Class for Table
    """

    def __init__(self, name: str, schema: Schema = Schema(), **kwargs):
        """
        :param name: table name
        :param schema: schema as defined by :class:`Schema`
        """
        if "." not in name:
            self.schema = schema
            self.raw_name = escape_identifier_name(name)
        else:
            schema_name, table_name = name.rsplit(".", 1)
            if len(schema_name.split(".")) > 2:
                # allow db.schema as schema_name, but a.b.c as schema_name is forbidden
                raise SQLLineageException("Invalid format for table name: %s.", name)
            self.schema = Schema(schema_name)
            self.raw_name = escape_identifier_name(table_name)
            if schema:
                warnings.warn("Name is in schema.table format, schema param is ignored")
        self.alias = escape_identifier_name(kwargs.pop("alias", self.raw_name))

    def __str__(self):
        return f"{self.schema}.{self.raw_name}"

    def __repr__(self):
        return "Table: " + str(self)

    def __eq__(self, other):
        return isinstance(other, Table) and str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    @staticmethod
    def of(table: Any) -> "Table":
        raise NotImplementedError


class Path:
    """
    Data Class for Path
    """

    def __init__(self, uri: str):
        """
        :param uri: uri of the path
        """
        self.uri = escape_identifier_name(uri)

    def __str__(self):
        return self.uri

    def __repr__(self):
        return "Path: " + str(self)

    def __eq__(self, other):
        return isinstance(other, Path) and self.uri == other.uri

    def __hash__(self):
        return hash(self.uri)


class SubQuery:
    """
    Data Class for SubQuery
    """

    def __init__(self, subquery: Any, subquery_raw: str, alias: Optional[str]):
        """
        :param subquery: subquery
        :param alias: subquery alias name
        """
        self.query = subquery
        self.query_raw = subquery_raw
        self.alias = (
            escape_identifier_name(alias)
            if alias is not None
            else f"subquery_{hash(self)}"
        )

    def __str__(self):
        return self.alias

    def __repr__(self):
        return "SubQuery: " + str(self)

    def __eq__(self, other):
        return isinstance(other, SubQuery) and self.query_raw == other.query_raw

    def __hash__(self):
        return hash(self.query_raw)

    @staticmethod
    def of(subquery: Any, alias: Optional[str]) -> "SubQuery":
        raise NotImplementedError


class Column:
    """
    Data Class for Column
    """

    def __init__(self, name: str, **kwargs):
        """
        :param name: column name
        :param parent: :class:`Table` or :class:`SubQuery`
        :param kwargs:
        """
        self._parent: Set[Union[Path, Table, SubQuery]] = set()
        self.raw_name = escape_identifier_name(name)
        self.source_columns = [
            (
                escape_identifier_name(raw_name),
                escape_identifier_name(qualifier) if qualifier is not None else None,
            )
            for raw_name, qualifier in kwargs.pop(
                "source_columns", ((self.raw_name, None),)
            )
        ]
        self.from_alias = kwargs.pop("from_alias", False)

    def __str__(self):
        return (
            f"{self.parent}.{self.raw_name}"
            if self.parent is not None and not isinstance(self.parent, Path)
            else f"{self.raw_name}"
        )

    def __repr__(self):
        return "Column: " + str(self)

    def __eq__(self, other):
        return (
            isinstance(other, Column)
            and str(self) == str(other)
            and self.parent == other.parent
        )

    def __hash__(self):
        return hash(str(self))

    @property
    def parent(self) -> Optional[Union[Path, Table, SubQuery]]:
        return next(iter(self._parent)) if len(self._parent) == 1 else None

    @parent.setter
    def parent(self, value: Union[Path, Table, SubQuery]):
        self._parent.add(value)

    @property
    def parent_candidates(self) -> List[Union[Path, Table, SubQuery]]:
        return sorted(self._parent, key=lambda p: str(p))

    @staticmethod
    def of(column: Any, **kwargs) -> "Column":
        """
        Build a 'Column' object
        :param column: column segment or token
        :return:
        """
        raise NotImplementedError

    def to_source_columns(self, alias_mapping: Dict[str, Union[Path, Table, SubQuery]]):
        """
        Best guess for source table given all the possible table/subquery and their alias.
        """

        def _to_src_col(
            name: str, parent: Optional[Union[Path, Table, SubQuery]] = None
        ) -> Column:
            col = Column(name)
            if parent:
                col.parent = parent
            return col

        source_columns = set()
        for src_col, qualifier in self.source_columns:
            if qualifier is None:
                if src_col == "*":
                    # select *
                    for table in set(alias_mapping.values()):
                        source_columns.add(_to_src_col(src_col, table))
                else:
                    # select unqualified column
                    source = _to_src_col(src_col, None)
                    for table in set(alias_mapping.values()):
                        # in case of only one table, we get the right answer
                        # in case of multiple tables, a bunch of possible tables are set
                        source.parent = table
                    source_columns.add(source)
            else:
                if alias_mapping.get(qualifier):
                    source_columns.add(
                        _to_src_col(src_col, alias_mapping.get(qualifier))
                    )
                else:
                    source_columns.add(_to_src_col(src_col, Table(qualifier)))
        return source_columns



from typing import Any, List, NamedTuple, Optional, Set, Union

from sqllineage.core.models import Column, SubQuery, Table


class SubQueryTuple(NamedTuple):
    parenthesis: Any
    alias: Optional[str]


class ColumnQualifierTuple(NamedTuple):
    column: str
    qualifier: Optional[str]


class AnalyzerContext(NamedTuple):
    # CTE queries that can be select from in current query context
    cte: Optional[Set[SubQuery]] = None
    # table that current top-level query is writing to, subquery in case of subquery context
    write: Optional[Set[Union[SubQuery, Table]]] = None
    # columns that write table specifies, used for `INSERT INTO x (col1, col2) SELECT` syntax
    write_columns: Optional[List[Column]] = None



from abc import abstractmethod
from typing import Dict, List

from sqllineage.core.models import Column, Table


class MetaDataProvider:
    """
    Base class used to provide metadata like table schema.

    When parse below sql:

    .. code-block:: sql

        INSERT INTO db1.table1
        SELECT c1
        FROM db2.table2 t2
        JOIN db3.table3 t3 ON t2.id = t3.id

    Only by literal analysis, we don't know which table is selected column c1 from.
    A subclass of MetaDataProvider implementing _get_table_columns passing to :class:`sqllineage.runner.LineageRunner`.
    can help parse column lineage correctly.
    """

    def __init__(self) -> None:
        self._session_metadata: Dict[str, List[str]] = {}

    def get_table_columns(self, table: Table, **kwargs) -> List[Column]:
        """
        return columns of given table.
        """
        if (key := str(table)) in self._session_metadata:
            cols = self._session_metadata[key]
        else:
            cols = self._get_table_columns(str(table.schema), table.raw_name, **kwargs)
        columns = []
        for col in cols:
            column = Column(col)
            column.parent = table
            columns.append(column)
        return columns

    @abstractmethod
    def _get_table_columns(self, schema: str, table: str, **kwargs) -> List[str]:
        """To be implemented by subclasses."""

    def register_session_metadata(self, table: Table, columns: List[Column]) -> None:
        """Register session-level metadata, like temporary table or view created."""
        self._session_metadata[str(table)] = [c.raw_name for c in columns]

    def deregister_session_metadata(self) -> None:
        """Deregister session-level metadata."""
        self._session_metadata.clear()

    def session(self):
        return MetaDataSession(self)

    def __bool__(self):
        """
        bool value tells whether this provider is ready to provide metadata
        """
        return True


class MetaDataSession:
    """
    Create an analyzer session which can register session-level metadata as a supplement to global metadata.
    This way, table or views created during the session can be queried.
    All session-level metadata will be deregistered once session closed.
    """

    def __init__(self, metadata_provider: MetaDataProvider):
        self.metadata_provider = metadata_provider

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.metadata_provider.deregister_session_metadata()

    def register_session_metadata(self, table: Table, columns: List[Column]) -> None:
        self.metadata_provider.register_session_metadata(table, columns)



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_select_column_using_case_when():
    sql = """INSERT INTO tab1
SELECT CASE WHEN col1 = 1 THEN 'V1' WHEN col1 = 2 THEN 'V2' END
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple(
                    "CASE WHEN col1 = 1 THEN 'V1' WHEN col1 = 2 THEN 'V2' END", "tab1"
                ),
            ),
        ],
    )
    sql = """INSERT INTO tab1
SELECT CASE WHEN col1 = 1 THEN 'V1' WHEN col1 = 2 THEN 'V2' END AS col2
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col2", "tab1"))],
    )


def test_select_column_using_case_when_with_subquery():
    sql = """INSERT INTO tab1
SELECT CASE WHEN (SELECT avg(col1) FROM tab3) > 0 AND col2 = 1 THEN (SELECT avg(col1) FROM tab3) ELSE 0 END AS col1
FROM tab4"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col2", "tab4"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col1", "tab3"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
        ],
    )


def test_select_column_using_multiple_case_when_with_subquery():
    sql = """INSERT INTO tab1
SELECT CASE
WHEN (SELECT avg(col1) FROM tab3) > 0 AND col2 = 1 THEN (SELECT avg(col1) FROM tab3)
WHEN (SELECT avg(col1) FROM tab3) > 0 AND col2 = 1 THEN (SELECT avg(col1) FROM tab5) ELSE 0 END AS col1
FROM tab4"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col2", "tab4"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col1", "tab3"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col1", "tab5"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
        ],
    )



import pytest

from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal, generate_metadata_providers


providers = generate_metadata_providers(
    {
        "db1.table1": ["id", "a", "b", "c", "d"],
        "db2.table2": ["id", "h", "i", "j", "k"],
        "db3.table3": ["pk", "p", "q", "r"],
    }
)


@pytest.mark.parametrize("provider", providers)
def test_select_column_from_tables(provider: MetaDataProvider):
    sql = """insert into db.tbl
select t1.id, a as x, b, h, i as y
from db1.table1 t1
join db2.table2 t2 on t1.id = t2.id
"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "db1.table1"),
                ColumnQualifierTuple("id", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("a", "db1.table1"),
                ColumnQualifierTuple("x", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("b", "db1.table1"),
                ColumnQualifierTuple("b", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("h", "db2.table2"),
                ColumnQualifierTuple("h", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("i", "db2.table2"),
                ColumnQualifierTuple("y", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )

    sql = """insert into db.tbl
    select a, b as x, h, i as y, p, q, pk as z
    from db1.table1 t1
    left join db2.table2 t2 on t1.id = t2.id
    right join db3.table3 t3 on t2.id = t3.pk
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db1.table1"),
                ColumnQualifierTuple("a", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("b", "db1.table1"),
                ColumnQualifierTuple("x", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("h", "db2.table2"),
                ColumnQualifierTuple("h", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("i", "db2.table2"),
                ColumnQualifierTuple("y", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("p", "db3.table3"),
                ColumnQualifierTuple("p", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("q", "db3.table3"),
                ColumnQualifierTuple("q", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("pk", "db3.table3"),
                ColumnQualifierTuple("z", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_column_from_subqueries(provider: MetaDataProvider):
    sql = """insert into db.tbl
select a, b as x, h, y
from (select a, b, c from db1.table1) t1
full join (select h, i, j as y from db2.table2) t2
on t1.id = t2.id
"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db1.table1"),
                ColumnQualifierTuple("a", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("b", "db1.table1"),
                ColumnQualifierTuple("x", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("h", "db2.table2"),
                ColumnQualifierTuple("h", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("j", "db2.table2"),
                ColumnQualifierTuple("y", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )

    sql = """insert into db.tbl
select a, b as x, h, y, pk as z, p
from (select id, a, b, c from db1.table1) t1
join (select id, h, i as y, j from db2.table2) t2 on t1.id = t2.id
left join (select pk, p, q from db3.table3) t3 on t1.id = t3.pk
"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db1.table1"),
                ColumnQualifierTuple("a", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("b", "db1.table1"),
                ColumnQualifierTuple("x", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("h", "db2.table2"),
                ColumnQualifierTuple("h", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("i", "db2.table2"),
                ColumnQualifierTuple("y", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("p", "db3.table3"),
                ColumnQualifierTuple("p", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("pk", "db3.table3"),
                ColumnQualifierTuple("z", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_column_from_table_subquery(provider: MetaDataProvider):
    sql = """insert into db.tbl
select a as x, b, q, pk as y
from db1.table1 t1
right join (select pk, p, q from db3.table3) t3
on t1.id = t3.pk
"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db1.table1"),
                ColumnQualifierTuple("x", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("b", "db1.table1"),
                ColumnQualifierTuple("b", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("pk", "db3.table3"),
                ColumnQualifierTuple("y", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("q", "db3.table3"),
                ColumnQualifierTuple("q", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )

    sql = """insert into db.tbl
select a, x, h as y, i, p, z
from (select id, a, b as x, c from db1.table1) t1
join db2.table2 t2 on t1.id = t2.id
left join (select pk, p, q as z from db3.table3) t3 on t2.id = t3.pk
"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db1.table1"),
                ColumnQualifierTuple("a", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("b", "db1.table1"),
                ColumnQualifierTuple("x", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("h", "db2.table2"),
                ColumnQualifierTuple("y", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("i", "db2.table2"),
                ColumnQualifierTuple("i", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("p", "db3.table3"),
                ColumnQualifierTuple("p", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("q", "db3.table3"),
                ColumnQualifierTuple("z", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_column_from_tempview_view_subquery(provider: MetaDataProvider):
    sql = """
create or replace view test_view
as
select id, a from db1.table1
;

insert into db.tbl
select test_view.id, a as x, h as y, i, p, z
from test_view
join db2.table2 t2 on test_view.id = t2.id
left join (select pk, p, q as z from db3.table3) t3 on t2.id = t3.pk
"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "db1.table1"),
                ColumnQualifierTuple("id", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("a", "db1.table1"),
                ColumnQualifierTuple("x", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("h", "db2.table2"),
                ColumnQualifierTuple("y", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("i", "db2.table2"),
                ColumnQualifierTuple("i", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("p", "db3.table3"),
                ColumnQualifierTuple("p", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("q", "db3.table3"),
                ColumnQualifierTuple("z", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_select_column():
    sql = """INSERT INTO tab1
SELECT col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT col1 AS col2
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col2", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT tab2.col1 AS col2
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col2", "tab1"))],
    )


def test_select_column_wildcard():
    sql = """INSERT INTO tab1
SELECT *
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("*", "tab2"), ColumnQualifierTuple("*", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT *
FROM tab2 a
         INNER JOIN tab3 b
                    ON a.id = b.id"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("*", "tab2"), ColumnQualifierTuple("*", "tab1")),
            (ColumnQualifierTuple("*", "tab3"), ColumnQualifierTuple("*", "tab1")),
        ],
    )


def test_select_column_wildcard_with_qualifier():
    sql = """INSERT INTO tab1
SELECT tab2.*
FROM tab2 a
         INNER JOIN tab3 b
                    ON a.id = b.id"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("*", "tab2"), ColumnQualifierTuple("*", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT a.*
FROM tab2 a
         INNER JOIN tab3 b
                    ON a.id = b.id"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("*", "tab2"), ColumnQualifierTuple("*", "tab1"))],
    )


def test_select_distinct_column():
    sql = """INSERT INTO tab1
SELECT DISTINCT col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_select_column_with_table_qualifier():
    sql = """INSERT INTO tab1
SELECT tab2.col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT t.col1
FROM tab2 AS t"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_select_column_with_table_qualifier_case_insensitive():
    sql = """INSERT INTO tab1
SELECT TA.col1
FROM tab2 AS ta"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT ta.col1
FROM tab2 AS TA"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """INSERT INTO tab1
SELECT Ta.col1
FROM tab2 AS tA"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_select_columns():
    sql = """INSERT INTO tab1
SELECT col1,
col2
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col2", "tab1"),
            ),
        ],
    )
    sql = """INSERT INTO tab1
SELECT max(col1),
max(col2)
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("max(col1)", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("max(col2)", "tab1"),
            ),
        ],
    )


def test_comment_after_column_comma_first():
    sql = """INSERT INTO tab1
SELECT a.col1
       --, a.col2
       , a.col3
FROM tab2 a"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col3", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
    )


def test_comment_after_column_comma_last():
    sql = """INSERT INTO tab1
SELECT a.col1,
       -- a.col2,
       a.col3
FROM tab2 a"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col3", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
    )


def test_select_column_name_case_insensitive():
    sql = """INSERT INTO tab1
SELECT col1,
COL1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            )
        ],
    )


def test_non_reserved_keyword_as_column_name():
    sql = """INSERT INTO tab1
SELECT host
FROM tab2 a"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("host", "tab2"),
                ColumnQualifierTuple("host", "tab1"),
            )
        ],
        test_sqlparse=False,
    )


def test_current_timestamp():
    """
    current_timestamp is a keyword since ANSI SQL-2016
    sqlparse cannot produce the correct AST for this case
    """
    sql = """INSERT INTO tab1
SELECT current_timestamp as col1,
       col2,
       col3
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col2", "tab2"),
                ColumnQualifierTuple("col2", "tab1"),
            ),
            (
                ColumnQualifierTuple("col3", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
        test_sqlparse=False,
    )



import itertools
from typing import Iterator, List, Union

from sqlparse.engine.grouping import _group, group_functions
from sqlparse.sql import (
    Case,
    Comment,
    Comparison,
    Function,
    Identifier,
    Parenthesis,
    TokenList,
    Values,
    Where,
)
from sqlparse.tokens import DML, Keyword, Name, Wildcard
from sqlparse.utils import recurse

from sqllineage.utils.entities import SubQueryTuple


def is_token_negligible(token: TokenList) -> bool:
    # utility to skip tokens like whitespace or comment
    return token.is_whitespace or isinstance(token, Comment)


def remove_parenthesis_between_union(token: Parenthesis) -> Parenthesis:
    """
    remove parenthesis around subqueries between union
    Example:
        INPUT: ((SELECT col1 FROM tab2) UNION ALL (SELECT col1 FROM tab3))
        OUTPUT: (SELECT col1 FROM tab2 UNION ALL SELECT col1 FROM tab3)
    """
    # pre-compute the idx of UNION/UNION ALL
    offsets = [-1]
    while True:
        offset, _ = token.token_next_by(
            m=[(Keyword, "UNION ALL"), (Keyword, "UNION")], idx=offsets[-1]
        )
        if offset is not None:
            offsets.append(offset)
        else:
            break
    expand_parenthesis_idx = set()
    for i, offset in enumerate(offsets[1:]):
        if i == 0:
            # first union idx, we need to handle the previous parenthesis
            pre_idx, sub_token = token.token_prev(idx=offset)
            if isinstance(sub_token, Parenthesis):
                expand_parenthesis_idx.add(pre_idx)
        # then all post parenthesis handled for each union
        post_idx, sub_token = token.token_next(idx=offset)
        if isinstance(sub_token, Parenthesis):
            expand_parenthesis_idx.add(post_idx)
    if expand_parenthesis_idx:
        updated_tokens = []
        for i, sub_token in enumerate(token.tokens):
            if i in expand_parenthesis_idx:
                inner_paren = get_innermost_parenthesis(sub_token)
                updated_tokens.extend(inner_paren.tokens[1:-1])
            else:
                updated_tokens.append(sub_token)
        token.tokens = updated_tokens
    return token


def get_innermost_parenthesis(token: Parenthesis) -> Parenthesis:
    # in case of subquery in nested parenthesis, find the innermost one first
    while True:
        idx, sub_paren = token.token_next_by(i=Parenthesis)
        if sub_paren is not None and (
            idx == 1
            or (idx > 1 and all(is_token_negligible(t) for t in token.tokens[1:idx]))
        ):
            # either the subsequent parenthesis is closely followed or there is only whitespace/comment in between
            token = sub_paren
        else:
            break
    return token


def is_subquery(token: TokenList) -> bool:
    flag = False
    if isinstance(token, Parenthesis):
        token = get_innermost_parenthesis(token)
        # check if innermost parenthesis contains SELECT
        _, sub_token = token.token_next_by(m=(DML, "SELECT"))
        if sub_token is not None:
            flag = True
    return flag


def is_values_clause(token: Parenthesis) -> bool:
    for t in token.tokens:
        if is_token_negligible(t):
            continue
        if t.match(Keyword, "VALUES"):
            return True
    return False


def get_subquery_parentheses(
    token: Union[Identifier, Function, Values, Where]
) -> List[SubQueryTuple]:
    """
    Retrieve subquery list
    the returned list is either empty when no subquery parsed or list of [parenthesis, alias] tuple
    """
    subquery = []
    as_idx, as_ = token.token_next_by(m=(Keyword, "AS"))
    sublist = list(token.get_sublists())
    if as_ is not None and len(sublist) == 1:
        # CTE: tbl AS (SELECT 1)
        target = sublist[0]
    else:
        if isinstance(token, Function):
            # CTE without AS: tbl (SELECT 1)
            target = token.tokens[-1]
        elif isinstance(token, (Values, Where)):
            # WHERE col1 IN (SELECT max(col1) FROM tab2)
            target = token
        else:
            # normal subquery: (SELECT 1) tbl
            target = token.token_first(skip_cm=True)
    if isinstance(target, (Case, Where)):
        # CASE WHEN (SELECT count(*) from tab1) > 0 THEN (SELECT count(*) FROM tab1) ELSE -1
        for tk in target.get_sublists():
            if isinstance(tk, Comparison):
                if is_subquery(tk.left):
                    subquery.append(SubQueryTuple(tk.left, tk.left.get_real_name()))
                if is_subquery(tk.right):
                    subquery.append(SubQueryTuple(tk.right, tk.right.get_real_name()))
            elif is_subquery(tk):
                subquery.append(SubQueryTuple(tk, token.get_real_name()))
    elif isinstance(target, Values):
        for row in target.get_sublists():
            for col in row:
                if is_subquery(col):
                    subquery.append(SubQueryTuple(col, col.get_real_name()))
    elif is_subquery(target):
        target = remove_parenthesis_between_union(target)
        subquery = [
            SubQueryTuple(get_innermost_parenthesis(target), token.get_real_name())
        ]
    return subquery


def get_parameters(token: Function):
    """
    This is a replacement for sqlparse.sql.Function.get_parameters(), which produces problematic result for:
        if(col1 = 'foo' AND col2 = 'bar', 1, 0)
    This implementation ignores the constant parameter as we don't need them for column lineage for now
    """
    if isinstance(token, Window):
        return token.get_parameters()
    else:
        return [
            tk for tk in token.tokens[-1].tokens if tk.is_group or tk.ttype == Wildcard
        ]


class Window(Function):
    """window function + OVER keyword + window defn"""

    def get_parameters(self) -> Iterator[TokenList]:
        return itertools.chain(
            get_parameters(self.get_window_function()),
            self.get_window_defn().get_sublists(),
        )

    def get_window_function(self) -> Function:
        return self.tokens[0]

    def get_window_defn(self) -> Parenthesis:
        return self.tokens[-1]


@recurse(Window)
def group_window(tlist):
    def match(token):
        return token.is_keyword and token.normalized == "OVER"

    def valid_prev(token):
        return isinstance(token, Function)

    def valid_next(token):
        return isinstance(token, Parenthesis)

    def post(tlist, pidx, tidx, nidx):
        return pidx, nidx

    _group(
        tlist, Window, match, valid_prev, valid_next, post, extend=False, recurse=False
    )


@recurse(Function)
def group_functions_as(tlist):
    """
    This function is to allow parsing columns in functions with CTAS syntax like:
        CREATE TABLE tbl1 AS SELECT coalesce(t1.col1, 0) AS col1 FROM t1;
    The code is mostly taken from original sqlparse.sql.group_functions by
    replacing with one condition.

    This is no longer needed if this PR get merged:
    https://github.com/andialbrecht/sqlparse/pull/662
    """
    has_create = False
    has_table = False
    has_as = False
    for tmp_token in tlist.tokens:
        if tmp_token.value == "CREATE":
            has_create = True
        if tmp_token.value == "TABLE":
            has_table = True
        if tmp_token.value == "AS":
            has_as = True
    if not has_create or not has_table or not has_as:
        return

    tidx, token = tlist.token_next_by(t=Name)
    while token:
        nidx, next_ = tlist.token_next(tidx)
        if isinstance(next_, Parenthesis):
            tlist.group_tokens(Function, tidx, nidx)
        tidx, token = tlist.token_next_by(t=Name, idx=tidx)


def group_function_with_window(tlist):
    group_functions(tlist)
    group_functions_as(tlist)
    group_window(tlist)



from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import StatementLineageHolder
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.core.parser.sqlfluff.utils import list_child_segments
from sqllineage.utils.entities import AnalyzerContext


class DropExtractor(BaseExtractor):
    """
    Drop statement lineage extractor
    """

    SUPPORTED_STMT_TYPES = ["drop_table_statement", "drop_view_statement"]

    def extract(
        self,
        statement: BaseSegment,
        context: AnalyzerContext,
    ) -> StatementLineageHolder:
        holder = StatementLineageHolder()
        drop_flag = False
        for segment in list_child_segments(statement):
            if (
                segment.type == "keyword"
                and segment.raw_upper in ["TABLE", "VIEW"]
                or (drop_flag is True and segment.raw_upper in ["IF", "EXISTS"])
            ):
                drop_flag = True
                continue
            if drop_flag:
                if table := self.find_table(segment):
                    holder.add_drop(table)
                drop_flag = False
        return holder



import os

import pytest

from sqllineage.core.metadata.sqlalchemy import SQLAlchemyMetaDataProvider
from sqllineage.exceptions import MetaDataProviderException


def test_sqlalchemy_metadata_provider_connection_fail():
    # connect to a directory as sqlite db, which is not possible. Simulate connection failure
    with pytest.raises(MetaDataProviderException):
        SQLAlchemyMetaDataProvider(f"sqlite:///{os.path.dirname(__file__)}")


def test_sqlalchemy_metadata_provider_driver_not_install():
    # use an unknown driver to connect. Simulate driver not installed
    with pytest.raises(MetaDataProviderException):
        SQLAlchemyMetaDataProvider("sqlite+unknown_driver:///:memory:")


def test_sqlalchemy_metadata_provider_query_fail():
    provider = SQLAlchemyMetaDataProvider("sqlite:///:memory:")
    assert (
        provider._get_table_columns("non_existing_schema", "non_existing_table") == []
    )
    assert provider._get_table_columns("main", "non_existing_table") == []



import logging
from argparse import Namespace
from typing import List

logger = logging.getLogger(__name__)


def escape_identifier_name(name: str):
    """
    conform to ANSI SQL standard that:
        1) unquoted identifier name is case-insensitive, convert to lower case
        2) quoted identifier name is case-sensitive, reserve case and remove quote char
    Reference: https://stackoverflow.com/a/19933159
    """
    quote_chars = ["`", '"', "'"]
    if any(quote_char in name for quote_char in quote_chars):
        for quote_char in quote_chars:
            name = name.strip(quote_char)
        return name
    else:
        return name.lower()


def extract_sql_from_args(args: Namespace) -> str:
    sql = ""
    if getattr(args, "f", None):
        try:
            with open(args.f) as f:
                sql = f.read()
        except IsADirectoryError:
            logger.exception("%s is a directory", args.f)
            exit(1)
        except FileNotFoundError:
            logger.exception("No such file: %s", args.f)
            exit(1)
        except PermissionError:
            # On Windows, open a directory as file throws PermissionError
            logger.exception("Permission denied when reading file '%s'", args.f)
            exit(1)
    elif getattr(args, "e", None):
        sql = args.e
    return sql


def split(sql: str) -> List[str]:
    # TODO: we need a parser independent split function
    import sqlparse
    from sqlparse.tokens import Punctuation

    result = []
    for s in sqlparse.parse(sql):
        if first_token := s.token_first(skip_cm=True):
            # sometimes sqlparse split out a statement that is comment only or semicolon only, we want to exclude that
            if first_token.ttype == Punctuation and first_token.value == ";":
                # exclude semicolon only statement
                continue
            else:
                result.append(s.value)
    return result


def trim_comment(sql: str) -> str:
    # TODO: we need a parser independent trim_comment function
    import sqlparse

    return str(sqlparse.format(sql, strip_comments=True))



from typing import Dict, List, Tuple, Union

from sqllineage.config import SQLLineageConfig
from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.models import Column, Path, SubQuery, Table
from sqllineage.exceptions import SQLLineageException


class SourceHandlerMixin:
    tables: List[Union[Path, SubQuery, Table]]
    columns: List[Column]
    union_barriers: List[Tuple[int, int]]

    def end_of_query_cleanup(self, holder: SubQueryLineageHolder) -> None:
        for i, tbl in enumerate(self.tables):
            holder.add_read(tbl)
        self.union_barriers.append((len(self.columns), len(self.tables)))
        for i, (col_barrier, tbl_barrier) in enumerate(self.union_barriers):
            prev_col_barrier, prev_tbl_barrier = (
                (0, 0) if i == 0 else self.union_barriers[i - 1]
            )
            col_grp = self.columns[prev_col_barrier:col_barrier]
            tbl_grp = self.tables[prev_tbl_barrier:tbl_barrier]
            if holder.write:
                if len(holder.write) > 1:
                    raise SQLLineageException
                tgt_tbl = next(iter(holder.write))
                lateral_column_aliases: Dict[str, List[Column]] = {}
                for idx, tgt_col_from_query in enumerate(col_grp):
                    tgt_col_from_query.parent = tgt_tbl
                    tgt_col_resolved = tgt_col_from_query
                    src_cols_resolved = []
                    for src_col in tgt_col_from_query.to_source_columns(
                        holder.get_alias_mapping_from_table_group(tbl_grp),
                    ):
                        if len(write_columns := holder.write_columns) == len(col_grp):
                            # example query: create view test (col3) select col1 as col2 from tab
                            # without write_columns = [col3] information, by default src_col = col1 and tgt_col = col2
                            # when write_columns exist and length matches, we want tgt_col = col3 instead of col2
                            # for invalid query: create view test (col3, col4) select col1 as col2 from tab,
                            # when the length doesn't match, we fall back to default behavior
                            tgt_col_resolved = write_columns[idx]
                        # lateral column alias handling
                        lca_flag = False
                        if SQLLineageConfig.LATERAL_COLUMN_ALIAS_REFERENCE:
                            if metadata_provider := getattr(
                                self, "metadata_provider", None
                            ):
                                from_dataset = False
                                for parent_candidate in src_col.parent_candidates:
                                    if isinstance(
                                        parent_candidate, Table
                                    ) and src_col in metadata_provider.get_table_columns(
                                        parent_candidate
                                    ):
                                        from_dataset = True
                                    elif isinstance(
                                        parent_candidate, SubQuery
                                    ) and src_col in holder.get_table_columns(
                                        parent_candidate
                                    ):
                                        from_dataset = True
                                if not from_dataset and (
                                    lca_cols_resolved := lateral_column_aliases.get(
                                        src_col.raw_name, []
                                    )
                                ):
                                    src_cols_resolved.extend(lca_cols_resolved)
                                    lca_flag = True
                        if not lca_flag:
                            src_cols_resolved.append(src_col)

                    for src_col_resolved in src_cols_resolved:
                        holder.add_column_lineage(src_col_resolved, tgt_col_resolved)
                        if (
                            SQLLineageConfig.LATERAL_COLUMN_ALIAS_REFERENCE
                            and tgt_col_from_query.from_alias
                        ):
                            lateral_column_aliases[tgt_col_from_query.raw_name] = (
                                src_cols_resolved
                            )



from ...helpers import assert_table_lineage_equal


def test_create():
    assert_table_lineage_equal(
        "CREATE TABLE tab1 (col1 STRING)",
        None,
        {"tab1"},
    )


def test_create_if_not_exist():
    assert_table_lineage_equal(
        "CREATE TABLE IF NOT EXISTS tab1 (col1 STRING)",
        None,
        {"tab1"},
    )


def test_create_as():
    assert_table_lineage_equal(
        "CREATE TABLE tab1 AS SELECT * FROM tab2", {"tab2"}, {"tab1"}
    )


def test_create_like():
    assert_table_lineage_equal("CREATE TABLE tab1 LIKE tab2", {"tab2"}, {"tab1"})


def test_create_view():
    assert_table_lineage_equal(
        """CREATE VIEW view1
as
SELECT
    col1,
    col2
FROM tab1
GROUP BY
col1""",
        {"tab1"},
        {"view1"},
    )


def test_create_as_with_parenthesis_around_select_statement():
    sql = "CREATE TABLE tab1 AS (SELECT * FROM tab2)"
    assert_table_lineage_equal(sql, {"tab2"}, {"tab1"}, test_sqlparse=False)


def test_create_as_with_parenthesis_around_both():
    sql = "CREATE TABLE tab1 AS (SELECT * FROM (tab2))"
    assert_table_lineage_equal(sql, {"tab2"}, {"tab1"}, test_sqlparse=False)



import pytest

from ...helpers import assert_table_lineage_equal


"""
This test class will contain all the tests for testing 'CTE Queries' where the dialect is not ANSI.
"""


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_with_insert_plus_table_keyword(dialect: str):
    assert_table_lineage_equal(
        "WITH tab1 AS (SELECT * FROM tab2) INSERT INTO TABLE tab3 SELECT * FROM tab1",
        {"tab2"},
        {"tab3"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_with_insert_overwrite(dialect: str):
    assert_table_lineage_equal(
        "WITH tab1 AS (SELECT * FROM tab2) INSERT OVERWRITE TABLE tab3 SELECT * FROM tab1",
        {"tab2"},
        {"tab3"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_with_insert_overwrite_without_table_keyword(dialect: str):
    assert_table_lineage_equal(
        "WITH tab1 AS (SELECT * FROM tab2) INSERT OVERWRITE tab3 SELECT * FROM tab1",
        {"tab2"},
        {"tab3"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_with_select_one_without_as(dialect: str):
    # AS in CTE is negligible in SparkSQL, however it is required in most other dialects
    # https://spark.apache.org/docs/latest/sql-ref-syntax-qry-select-cte.html
    # https://dev.mysql.com/doc/refman/8.0/en/with.html
    assert_table_lineage_equal(
        "WITH wtab1 (SELECT * FROM schema1.tab1) SELECT * FROM wtab1",
        {"schema1.tab1"},
        dialect=dialect,
    )



import os
from unittest.mock import patch

import pytest

from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal, generate_metadata_providers

providers = generate_metadata_providers(
    {
        "public.src_tbl1": ["id", "a", "b", "c"],
    }
)


@pytest.mark.parametrize("provider", providers)
@patch.dict(os.environ, {"SQLLINEAGE_LATERAL_COLUMN_ALIAS_REFERENCE": "1"})
def test_column_top_level_enable_lateral_ref(
    provider: MetaDataProvider,
):
    sql = """
    insert into public.tgt_tbl1
    select
        name               as user_name,
        user_name || email as id
    from
        public.src_tbl1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("user_name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )
    sql = """
    insert into public.tgt_tbl1
    (
        name,
        email
    )
    select
        st1.name,
        st1.name || st1.email || '@gmail.com' as email
    from
        public.src_tbl1 as st1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl1"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )
    sql = """
    insert into public.tgt_tbl1
    (
        id,
        id_original
    )
    select
        'a || b || c' || id as id,
        id                  as id_original
    from
        public.src_tbl1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id_original", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
@patch.dict(os.environ, {"SQLLINEAGE_LATERAL_COLUMN_ALIAS_REFERENCE": "1"})
def test_column_top_level_enable_lateral_ref_with_metadata_from_table(
    provider: MetaDataProvider,
):
    sql = """
    insert into public.tgt_tbl1
    select
        a || b || c || id as id,
        id                as id_original
    from
        public.src_tbl1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("b", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("c", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id_original", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
@patch.dict(os.environ, {"SQLLINEAGE_LATERAL_COLUMN_ALIAS_REFERENCE": "1"})
def test_column_top_level_enable_lateral_ref_with_metadata_from_subquery(
    provider: MetaDataProvider,
):
    sql = """
    insert into public.tgt_tbl1
    select
        a || b || c || id as id,
        id                as id_original
    from
        (
            select
               *
            from
                public.src_tbl1
        ) as sq
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("b", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("c", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id_original", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
@patch.dict(os.environ, {"SQLLINEAGE_LATERAL_COLUMN_ALIAS_REFERENCE": "1"})
def test_column_top_level_enable_lateral_ref_with_metadata_from_nested_subquery(
    provider: MetaDataProvider,
):
    sql = """
    insert into public.tgt_tbl1
    select
        a || b || c || id as id,
        id                as id_original
    from
        (
            select
                *
            from
                (
                    select
                       *
                    from
                        public.src_tbl1
                ) as inner_sq
        ) as outer_sq
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("b", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("c", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id_original", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
@patch.dict(os.environ, {"SQLLINEAGE_LATERAL_COLUMN_ALIAS_REFERENCE": "1"})
def test_column_enable_lateral_ref_within_subquery(
    provider: MetaDataProvider,
):
    sql = """
    insert into public.tgt_tbl1
    select
        sq.name
    from
        (
            select
                id     || a as alias1,
                alias1 || b as name
            from
                public.src_tbl1
        ) as sq
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("a", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("b", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )

    sql = """
    insert into public.tgt_tbl1
    select
        sq.name
    from
        (
            select
                st1.id || st1.name as alias1,
                alias1 || st2.email as name
            from
                public.src_tbl1 as st1
                join
                    public.src_tbl2 as st2
                on
                    st1.id = st2.id
        ) as sq
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl2"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
        metadata_provider=provider,
    )


def test_column_top_level_disable_lateral_ref():
    sql = """
    insert into public.tgt_tbl1
    select
        name               as user_name,
        user_name || email as id
    from
        public.src_tbl1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("user_name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("user_name", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )
    sql = """
    insert into public.tgt_tbl1
    (
        name,
        email
    )
    select
        st1.name,
        st1.name || st1.email || '@gmail.com' as email
    from
        public.src_tbl1 as st1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("name", "public.src_tbl1"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl1"),
                ColumnQualifierTuple("email", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )
    sql = """
    insert into public.tgt_tbl1
    (
        id,
        id_original
    )
    select
        'a || b || c' || id as id,
        id                  as id_original
    from
        public.src_tbl1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id_original", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )


def test_column_top_level_disable_lateral_ref_with_metadata_from_table():
    sql = """
    insert into public.tgt_tbl1
    select
        a || b || c || id as id,
        id                as id_original
    from
        public.src_tbl1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("b", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("c", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("id", "public.src_tbl1"),
                ColumnQualifierTuple("id_original", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )


def test_column_disable_lateral_ref_within_subquery():
    sql = """
    insert into public.tgt_tbl1
    select
        sq.name
    from
        (
            select
                id     || a as alias1,
                alias1 || b as name
            from
                public.src_tbl1
        ) as sq
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("alias1", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("b", "public.src_tbl1"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )

    sql = """
    insert into public.tgt_tbl1
    select
        sq.name
    from
        (
            select
                st1.id || st1.name as alias1,
                alias1 || st2.email as name
            from
                public.src_tbl1 as st1
                join
                    public.src_tbl2 as st2
                on
                    st1.id = st2.id
        ) as sq
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("alias1", None),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
            (
                ColumnQualifierTuple("email", "public.src_tbl2"),
                ColumnQualifierTuple("name", "public.tgt_tbl1"),
            ),
        ],
        test_sqlparse=False,
    )



import logging
from typing import Any, Dict, List, Optional

from sqlalchemy import MetaData, Table, create_engine, make_url
from sqlalchemy.exc import NoSuchModuleError, NoSuchTableError, OperationalError

from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.exceptions import MetaDataProviderException


logger = logging.getLogger(__name__)


class SQLAlchemyMetaDataProvider(MetaDataProvider):
    """
    SQLAlchemyMetaDataProvider queries metadata from database using SQLAlchemy
    """

    def __init__(self, url: str, engine_kwargs: Optional[Dict[str, Any]] = None):
        """
        :param url: sqlalchemy url
        :param engine_kwargs: a dictionary of keyword arguments that will be passed to sqlalchemy create_engine
        """
        super().__init__()
        self.metadata_obj = MetaData()
        try:
            if engine_kwargs is None:
                engine_kwargs = {}
            self.engine = create_engine(url, **engine_kwargs)
        except NoSuchModuleError as e:
            u = make_url(url)
            raise MetaDataProviderException(
                f"SQLAlchemy dialect driver {u.drivername} is not installed correctly"
            ) from e
        try:
            self.engine.connect()
        except OperationalError as e:
            raise MetaDataProviderException(f"Could not connect to {url}") from e

    def _get_table_columns(self, schema: str, table: str, **kwargs) -> List[str]:
        columns = []
        try:
            sqlalchemy_table = Table(
                table, self.metadata_obj, schema=schema, autoload_with=self.engine
            )
            columns = [c.name for c in sqlalchemy_table.columns]
        except (NoSuchTableError, OperationalError):
            logger.warning(
                "error listing columns for table %s.%s in %s, return empty list instead",
                schema,
                table,
                self.engine.url,
            )
        return columns



import pytest

from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal, generate_metadata_providers


providers = generate_metadata_providers(
    {
        "db.tbl_x": ["id", "a", "b"],
        "db.tbl_y": ["id", "h", "i"],
        "db.tbl_z": ["pk", "s", "t"],
    }
)


@pytest.mark.parametrize("provider", providers)
def test_select_single_table_wildcard(provider: MetaDataProvider):
    sql = """insert into test_v
    select *
    from db.tbl_x
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "db.tbl_x"),
                ColumnQualifierTuple("id", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("b", "db.tbl_x"),
                ColumnQualifierTuple("b", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_single_table_wildcard_in_subquery(provider: MetaDataProvider):
    sql = """insert into test_v
    select id, b
    from (
        select *
        from (
            select a, b, id from db.tbl_x
        ) t1
    ) t2
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "db.tbl_x"),
                ColumnQualifierTuple("id", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("b", "db.tbl_x"),
                ColumnQualifierTuple("b", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_single_table_partial_wildcard_in_subquery(provider: MetaDataProvider):
    sql = """insert into test_v
    select a, b
    from (
        select *, row_number() over (partition by id) as rn
        from db.tbl_x t
    ) t1
    where rn = 1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("b", "db.tbl_x"),
                ColumnQualifierTuple("b", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )
    sql = """insert into test_v
    select a, b
    from (
        select *, row_number() over (partition by a) as rn
        from db.tbl_x t
    ) t1
    where rn = 1
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("b", "db.tbl_x"),
                ColumnQualifierTuple("b", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_table_join_partial_wildcard_at_last(provider: MetaDataProvider):
    sql = """insert into test_v
    select a, id, h
    from (
        select x.a, y.* from db.tbl_x x
        join db.tbl_y y on x.id = y.id
    ) t
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("id", "db.tbl_y"),
                ColumnQualifierTuple("id", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("h", "db.tbl_y"),
                ColumnQualifierTuple("h", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_table_join_partial_wildcard_at_beginning(provider: MetaDataProvider):
    sql = """insert into test_v
    select a, id, h
    from (
        select x.*, y.h from db.tbl_x x
        join db.tbl_y y on x.id = y.id
    ) t
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("id", "db.tbl_x"),
                ColumnQualifierTuple("id", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("h", "db.tbl_y"),
                ColumnQualifierTuple("h", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_table_join_multiple_wildcards(provider: MetaDataProvider):
    sql = """insert into test_v
    select a, h
    from (
        select x.*, y.* from db.tbl_x x
        join db.tbl_y y on x.id = y.id
    ) t
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("h", "db.tbl_y"),
                ColumnQualifierTuple("h", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_table_join_multiple_wildcards_merged_at_top_level(
    provider: MetaDataProvider,
):
    sql = """insert into test_v
    select *
    from (
        select x.*, z.* from db.tbl_x x
        join db.tbl_z z on x.id = z.pk
    ) t
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("b", "db.tbl_x"),
                ColumnQualifierTuple("b", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("id", "db.tbl_x"),
                ColumnQualifierTuple("id", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("pk", "db.tbl_z"),
                ColumnQualifierTuple("pk", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("s", "db.tbl_z"),
                ColumnQualifierTuple("s", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("t", "db.tbl_z"),
                ColumnQualifierTuple("t", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_select_table_join_multiple_wildcards_at_different_level(
    provider: MetaDataProvider,
):
    sql = """insert into test_v
    select x.*, y.h, y.i
    from (select * from db.tbl_x where a = 0) x
    join (select *
          from (
              select *, row_number() over (partition by i) as rn from db.tbl_y
          ) t1
          where rn = 1
    ) y
    on x.id = y.id
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("b", "db.tbl_x"),
                ColumnQualifierTuple("b", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("id", "db.tbl_x"),
                ColumnQualifierTuple("id", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("h", "db.tbl_y"),
                ColumnQualifierTuple("h", "<default>.test_v"),
            ),
            (
                ColumnQualifierTuple("i", "db.tbl_y"),
                ColumnQualifierTuple("i", "<default>.test_v"),
            ),
        ],
        metadata_provider=provider,
    )


@pytest.mark.parametrize("provider", providers)
def test_wildcard_reference_from_previous_statements(provider: MetaDataProvider):
    sql = """create table test_x as
    select a, b
    from (
        select *, row_number() over (partition by id) as rn
        from db.tbl_x t
    ) t1
    where rn = 1
    ;

    create table test_z as
    select *
    from db.tbl_z
    ;

    insert into db.tbl
    select t1.*, t2.h, t3.*
    from test_x t1
    join (select * from db.tbl_y where i > 0) t2 on t1.id = t2.id
    left join test_z t3 on t2.id = t3.pk
    """
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("a", "db.tbl_x"),
                ColumnQualifierTuple("a", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("b", "db.tbl_x"),
                ColumnQualifierTuple("b", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("h", "db.tbl_y"),
                ColumnQualifierTuple("h", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("pk", "db.tbl_z"),
                ColumnQualifierTuple("pk", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("s", "db.tbl_z"),
                ColumnQualifierTuple("s", "db.tbl"),
            ),
            (
                ColumnQualifierTuple("t", "db.tbl_z"),
                ColumnQualifierTuple("t", "db.tbl"),
            ),
        ],
        metadata_provider=provider,
    )



from typing import List
from typing import Optional, Tuple

from sqlfluff.core.parser import BaseSegment

from sqllineage import SQLPARSE_DIALECT
from sqllineage.core.models import Column, Schema, SubQuery, Table
from sqllineage.core.parser.sqlfluff.utils import (
    extract_column_qualifier,
    extract_identifier,
    is_subquery,
    is_wildcard,
    list_child_segments,
)
from sqllineage.utils.entities import ColumnQualifierTuple
from sqllineage.utils.helpers import escape_identifier_name

NON_IDENTIFIER_OR_COLUMN_SEGMENT_TYPE = [
    "function",
    "over_clause",
    "partitionby_clause",
    "orderby_clause",
    "expression",
    "case_expression",
    "when_clause",
    "else_clause",
    "select_clause_element",
    "cast_expression",
]

SOURCE_COLUMN_SEGMENT_TYPE = NON_IDENTIFIER_OR_COLUMN_SEGMENT_TYPE + [
    "identifier",
    "column_reference",
]


class SqlFluffTable(Table):
    """
    Data Class for SqlFluffTable
    """

    @staticmethod
    def of(table: BaseSegment, alias: Optional[str] = None) -> Table:
        """
        Build an object of type 'Table'
        :param table: table segment to be processed
        :param alias: alias of the table segment
        :return: 'Table' object
        """
        dot_idx = None
        for idx in range(len(table.segments) - 2, -1, -1):
            token = table.segments[idx]
            if bool(token.type == "symbol"):
                dot_idx, _ = idx, token
                break
        real_name = (
            table.segments[dot_idx + 1].raw
            if dot_idx
            else (table.raw if table.type == "identifier" else table.segments[0].raw)
        )
        # rewrite identifier's get_parent_name accordingly
        parent_name = (
            "".join(
                [
                    escape_identifier_name(segment.raw)
                    for segment in table.segments[:dot_idx]
                ]
            )
            if dot_idx
            else None
        )
        schema = Schema(parent_name) if parent_name is not None else Schema()
        kwargs = {"alias": alias} if alias else {}
        return Table(real_name, schema, **kwargs)


class SqlFluffSubQuery(SubQuery):
    """
    Data Class for SqlFluffSubQuery
    """

    @staticmethod
    def of(subquery: BaseSegment, alias: Optional[str]) -> SubQuery:
        """
        Build a 'SubQuery' object
        :param subquery: subquery segment
        :param alias: subquery alias
        :return: 'SubQuery' object
        """
        return SubQuery(subquery, subquery.raw, alias)


class SqlFluffColumn(Column):
    """
    Data Class for SqlFluffColumn
    """

    @staticmethod
    def of(column: BaseSegment, **kwargs) -> Column:
        """
        Build a 'Column' object
        :param column: column segment
        :return: 'Column' object
        """
        if column.type == "select_clause_element":
            source_columns, alias = SqlFluffColumn._get_column_and_alias(column)
            if alias:
                return Column(alias, source_columns=source_columns, from_alias=True)
            if source_columns:
                column_name = None
                for sub_segment in list_child_segments(column):
                    if sub_segment.type == "column_reference" or is_wildcard(
                        sub_segment
                    ):
                        if cqt := extract_column_qualifier(sub_segment):
                            column_name = cqt.column
                    elif sub_segment.type == "expression":
                        # special handling for postgres style type cast, col as target column name instead of col::type
                        if len(sub2_segments := list_child_segments(sub_segment)) == 1:
                            if (
                                sub2_segment := sub2_segments[0]
                            ).type == "cast_expression":
                                if (
                                    len(
                                        sub3_segments := list_child_segments(
                                            sub2_segment
                                        )
                                    )
                                    == 2
                                ):
                                    if (
                                        sub3_segment := sub3_segments[0]
                                    ).type == "column_reference":
                                        if cqt := extract_column_qualifier(
                                            sub3_segment
                                        ):
                                            column_name = cqt.column
                return Column(
                    column.raw if column_name is None else column_name,
                    source_columns=source_columns,
                )

        # Wildcard, Case, Function without alias (thus not recognized as an Identifier)
        source_columns = SqlFluffColumn._extract_source_columns(column)
        return Column(column.raw, source_columns=source_columns)

    @staticmethod
    def _extract_source_columns(segment: BaseSegment) -> List[ColumnQualifierTuple]:
        """
        :param segment: segment to be processed
        :return: list of extracted source columns
        """
        col_list = []
        if segment.type in ("identifier", "column_reference") or is_wildcard(segment):
            if cqt := extract_column_qualifier(segment):
                col_list = [cqt]
        elif segment.type in NON_IDENTIFIER_OR_COLUMN_SEGMENT_TYPE:
            sub_segments = list_child_segments(segment)
            col_list = []
            for sub_segment in sub_segments:
                if sub_segment.type == "bracketed":
                    if is_subquery(sub_segment):
                        col_list += SqlFluffColumn._get_column_from_subquery(
                            sub_segment
                        )
                    else:
                        col_list += SqlFluffColumn._get_column_from_parenthesis(
                            sub_segment
                        )
                elif sub_segment.type in SOURCE_COLUMN_SEGMENT_TYPE or is_wildcard(
                    sub_segment
                ):
                    res = SqlFluffColumn._extract_source_columns(sub_segment)
                    col_list.extend(res)
        return col_list

    @staticmethod
    def _get_column_from_subquery(
        sub_segment: BaseSegment,
    ) -> List[ColumnQualifierTuple]:
        """
        :param sub_segment: segment to be processed
        :return: A list of source columns from a segment
        """
        # This is to avoid circular import
        from sqllineage.runner import LineageRunner

        src_cols = [
            lineage[0]
            for lineage in LineageRunner(
                sub_segment.raw, dialect=SQLPARSE_DIALECT
            ).get_column_lineage(exclude_path_ending_in_subquery=False)
        ]
        source_columns = [
            ColumnQualifierTuple(
                src_col.raw_name, src_col.parent.raw_name if src_col.parent else None
            )
            for src_col in src_cols
        ]
        return source_columns

    @staticmethod
    def _get_column_from_parenthesis(
        sub_segment: BaseSegment,
    ) -> List[ColumnQualifierTuple]:
        # windows function has an extra layer, get rid of it so that it can be handled as regular functions
        if window_specification := sub_segment.get_child("window_specification"):
            sub_segment = window_specification
        col, _ = SqlFluffColumn._get_column_and_alias(sub_segment, False)
        return col if col else []

    @staticmethod
    def _get_column_and_alias(
        segment: BaseSegment, check_bracketed: bool = True
    ) -> Tuple[List[ColumnQualifierTuple], Optional[str]]:
        """
        check_bracketed is True for top-level column definition, like (col1 + col2) as col3
        set to False for bracket in function call, like coalesce(col1, col2) as col3
        """
        alias = None
        columns = []
        sub_segments = list_child_segments(segment, check_bracketed)
        for sub_segment in sub_segments:
            if sub_segment.type == "alias_expression":
                alias = extract_identifier(sub_segment)
            elif sub_segment.type in SOURCE_COLUMN_SEGMENT_TYPE or is_wildcard(
                sub_segment
            ):
                res = SqlFluffColumn._extract_source_columns(sub_segment)
                columns += res if res else []
        return columns, alias



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_select_column_from_table_join():
    sql = """INSERT INTO tab1
SELECT tab2.col1,
       tab3.col2
FROM tab2
         INNER JOIN tab3
                    ON tab2.id = tab3.id"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab3"),
                ColumnQualifierTuple("col2", "tab1"),
            ),
        ],
    )
    sql = """INSERT INTO tab1
SELECT tab2.col1 AS col3,
       tab3.col2 AS col4
FROM tab2
         INNER JOIN tab3
                    ON tab2.id = tab3.id"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab3"),
                ColumnQualifierTuple("col4", "tab1"),
            ),
        ],
    )
    sql = """INSERT INTO tab1
SELECT a.col1 AS col3,
       b.col2 AS col4
FROM tab2 a
         INNER JOIN tab3 b
                    ON a.id = b.id"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
            (
                ColumnQualifierTuple("col2", "tab3"),
                ColumnQualifierTuple("col4", "tab1"),
            ),
        ],
    )


def test_select_column_without_table_qualifier_from_table_join():
    sql = """INSERT INTO tab3
SELECT f1
FROM ( SELECT f1 FROM tab1)
LEFT JOIN ( SELECT f1 FROM tab2) USING (f1)"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("f1", "tab1"),
                ColumnQualifierTuple("f1", "tab3"),
            ),
            (
                ColumnQualifierTuple("f1", "tab2"),
                ColumnQualifierTuple("f1", "tab3"),
            ),
        ],
    )


def test_select_column_from_same_table_multiple_time_using_different_alias():
    sql = """INSERT INTO tab1
SELECT a.col1 AS col2,
       b.col1 AS col3
FROM tab2 a
         JOIN tab2 b
              ON a.parent_id = b.id"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col2", "tab1"),
            ),
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col3", "tab1"),
            ),
        ],
    )


def test_column_reference_with_ansi89_join():
    sql = """INSERT INTO tab3
SELECT a.id,
       a.name AS name1,
       b.name AS name2
FROM (SELECT id, name
      FROM tab1) a,
     (SELECT id, name
      FROM tab2) b
WHERE a.id = b.id"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("id", "tab1"), ColumnQualifierTuple("id", "tab3")),
            (
                ColumnQualifierTuple("name", "tab1"),
                ColumnQualifierTuple("name1", "tab3"),
            ),
            (
                ColumnQualifierTuple("name", "tab2"),
                ColumnQualifierTuple("name2", "tab3"),
            ),
        ],
    )


def test_column_lineage_multiple_paths_for_same_column():
    sql = """INSERT INTO tab2
SELECT tab1.id,
       coalesce(join_table_1.col1, join_table_2.col1, join_table_3.col1) AS col1
FROM tab1
         LEFT JOIN (SELECT id, col1 FROM tab1 WHERE flag = 1) AS join_table_1
                   ON tab1.id = join_table_1.id
         LEFT JOIN (SELECT id, col1 FROM tab1 WHERE flag = 2) AS join_table_2
                   ON tab1.id = join_table_2.id
         LEFT JOIN (SELECT id, col1 FROM tab1 WHERE flag = 3) AS join_table_3
                   ON tab1.id = join_table_3.id"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "tab1"),
                ColumnQualifierTuple("id", "tab2"),
            ),
            (
                ColumnQualifierTuple("col1", "tab1"),
                ColumnQualifierTuple("col1", "tab2"),
            ),
        ],
    )



import pytest

from ...helpers import assert_table_lineage_equal


"""
This test class will contain all the tests for testing 'Insert Queries' where the dialect is not ANSI.
"""


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_insert_overwrite(dialect: str):
    assert_table_lineage_equal(
        "INSERT OVERWRITE TABLE tab1 SELECT col1 FROM tab2",
        {"tab2"},
        {"tab1"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_insert_overwrite_from_self(dialect: str):
    assert_table_lineage_equal(
        """INSERT OVERWRITE TABLE foo
SELECT col FROM foo
WHERE flag IS NOT NULL""",
        {"foo"},
        {"foo"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_insert_overwrite_from_self_with_join(dialect: str):
    assert_table_lineage_equal(
        """INSERT OVERWRITE TABLE tab_1
SELECT tab_2.col_a from tab_2
JOIN tab_1
ON tab_1.col_a = tab_2.cola""",
        {"tab_1", "tab_2"},
        {"tab_1"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_insert_overwrite_values(dialect: str):
    assert_table_lineage_equal(
        "INSERT OVERWRITE TABLE tab1 VALUES ('val1', 'val2'), ('val3', 'val4')",
        {},
        {"tab1"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_insert_into_with_keyword_table(dialect: str):
    assert_table_lineage_equal(
        "INSERT INTO TABLE tab1 VALUES (1, 2)", set(), {"tab1"}, dialect=dialect
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_insert_into_partitions(dialect: str):
    assert_table_lineage_equal(
        "INSERT INTO TABLE tab1 PARTITION (par1=1) SELECT * FROM tab2",
        {"tab2"},
        {"tab1"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_insert_overwrite_without_table_keyword(dialect: str):
    assert_table_lineage_equal(
        "INSERT OVERWRITE tab1 SELECT * FROM tab2",
        {"tab2"},
        {"tab1"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_lateral_view_using_json_tuple(dialect: str):
    sql = """INSERT OVERWRITE TABLE foo
SELECT sc.id, q.item0, q.item1
FROM bar sc
LATERAL VIEW json_tuple(sc.json, 'key1', 'key2') q AS item0, item1"""
    assert_table_lineage_equal(sql, {"bar"}, {"foo"}, dialect)


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_lateral_view_outer(dialect: str):
    sql = """INSERT OVERWRITE TABLE foo
SELECT sc.id, q.col1
FROM bar sc
LATERAL VIEW OUTER explode(sc.json_array) q AS col1"""
    assert_table_lineage_equal(sql, {"bar"}, {"foo"}, dialect)



from abc import abstractmethod
from typing import List

from sqllineage.core.holders import StatementLineageHolder
from sqllineage.core.metadata_provider import MetaDataProvider


class LineageAnalyzer:
    """SQL Statement Level Lineage Analyzer
    Parser specific implementation should inherit this class and implement analyze method
    """

    PARSER_NAME: str = ""
    SUPPORTED_DIALECTS: List[str] = []

    @abstractmethod
    def analyze(
        self, sql: str, metadata_provider: MetaDataProvider
    ) -> StatementLineageHolder:
        """
        to analyze single statement sql and store the result into StatementLineageHolder.

        :param sql: single-statement SQL string to be processed
        :param metadata_provider: :class:`sqllineage.core.metadata_provider.MetaDataProvider` provides metadata on
                                  tables to help lineage analyzing
        :return: :class:`sqllineage.core.holders.StatementLineageHolder`
        """



import pytest

from ...helpers import assert_table_lineage_equal


@pytest.mark.parametrize("dialect", ["bigquery"])
def test_merge_without_into(dialect: str):
    """
    INTO is optional in BigQuery MERGE statement:
    https://cloud.google.com/bigquery/docs/reference/standard-sql/dml-syntax#merge_statement
    """
    sql = """MERGE target USING src AS b ON target.k = b.k
WHEN MATCHED THEN UPDATE SET target.v = b.v
WHEN NOT MATCHED THEN INSERT (k, v) VALUES (b.k, b.v)"""
    assert_table_lineage_equal(sql, {"src"}, {"target"}, dialect=dialect)


@pytest.mark.parametrize("dialect", ["bigquery"])
def test_merge_insert_row(dialect: str):
    """
    MERGE INSERT CLAUSE in BigQuery can be INSERT ROW without specifying columns via INSERT VALUES (col, ...)
    https://cloud.google.com/bigquery/docs/reference/standard-sql/dml-syntax#merge_statement
    """
    sql = """MERGE INTO tgt t
USING src s
ON t.date = s.date and t.channel = s.channel
WHEN NOT MATCHED THEN
INSERT ROW
WHEN MATCHED THEN
UPDATE SET t.col = s.col"""
    assert_table_lineage_equal(
        sql,
        {"src"},
        {"tgt"},
        dialect=dialect,
    )



import pytest

from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


@pytest.mark.parametrize("dialect", ["tsql"])
def test_tsql_assignment_operator(dialect: str):
    """
    Assignment Operator is a Transact-SQL specific feature, used interchangeably with column alias
    https://learn.microsoft.com/en-us/sql/t-sql/language-elements/assignment-operator-transact-sql?view=sql-server-ver15
    """
    sql = """INSERT INTO foo
SELECT FirstColumnHeading = 'xyz',
       SecondColumnHeading = ProductID
FROM Production.Product"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("ProductID", "Production.Product"),
                ColumnQualifierTuple("SecondColumnHeading", "foo"),
            )
        ],
        dialect=dialect,
        test_sqlparse=False,
    )



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_column_reference_from_cte_using_qualifier():
    sql = """WITH wtab1 AS (SELECT col1 FROM tab2)
INSERT INTO tab1
SELECT wtab1.col1 FROM wtab1"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_column_reference_from_cte_using_alias():
    sql = """WITH wtab1 AS (SELECT col1 FROM tab2)
INSERT INTO tab1
SELECT wt.col1 FROM wtab1 wt"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )


def test_column_reference_from_previous_defined_cte():
    sql = """WITH
cte1 AS (SELECT a FROM tab1),
cte2 AS (SELECT a FROM cte1)
INSERT INTO tab2
SELECT a FROM cte2"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("a", "tab1"), ColumnQualifierTuple("a", "tab2"))],
    )


def test_multiple_column_references_from_previous_defined_cte():
    sql = """WITH
cte1 AS (SELECT a, b FROM tab1),
cte2 AS (SELECT a, max(b) AS b_max, count(b) AS b_cnt FROM cte1 GROUP BY a)
INSERT INTO tab2
SELECT cte1.a, cte2.b_max, cte2.b_cnt FROM cte1 JOIN cte2
WHERE cte1.a = cte2.a"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("a", "tab1"), ColumnQualifierTuple("a", "tab2")),
            (ColumnQualifierTuple("b", "tab1"), ColumnQualifierTuple("b_max", "tab2")),
            (ColumnQualifierTuple("b", "tab1"), ColumnQualifierTuple("b_cnt", "tab2")),
        ],
    )


def test_column_reference_from_cte_and_union():
    sql = """WITH cte_1 AS (select col1 from tab1),
cte_2 AS (SELECT col1 from tab2)
INSERT INTO tab3
    SELECT col1 from cte_1
    UNION
    SELECT col1 from cte_2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab1"),
                ColumnQualifierTuple("col1", "tab3"),
            ),
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab3"),
            ),
        ],
    )


def test_smarter_column_resolution_using_query_context():
    sql = """WITH
cte1 AS (SELECT a, b FROM tab1),
cte2 AS (SELECT c, d FROM tab2)
INSERT INTO tab3
SELECT b, d FROM cte1 JOIN cte2
WHERE cte1.a = cte2.c"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("b", "tab1"), ColumnQualifierTuple("b", "tab3")),
            (ColumnQualifierTuple("d", "tab2"), ColumnQualifierTuple("d", "tab3")),
        ],
    )


def test_cte_inside_insert_in_parenthesis():
    sql = """INSERT INTO tab3 (WITH tab1 AS (SELECT * FROM tab2) SELECT * FROM tab1)"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("*", "tab2"),
                ColumnQualifierTuple("*", "tab3"),
            ),
        ],
        test_sqlparse=False,
    )


def test_select_column_from_cte_case_insensitive():
    sql = """WITH CTE1 AS (SELECT col1 FROM tab2)
INSERT INTO tab1
SELECT col1 FROM cte1"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """WITH cte1 AS (SELECT col1 FROM tab2)
INSERT INTO tab1
SELECT col1 FROM CTE1"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )
    sql = """WITH CTe1 AS (SELECT col1 FROM tab2)
INSERT INTO tab1
SELECT col1 FROM cTE1"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("col1", "tab2"), ColumnQualifierTuple("col1", "tab1"))],
    )



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_merge_into_update():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN MATCHED THEN UPDATE SET target.v = src.v"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("v", "src"), ColumnQualifierTuple("v", "target"))],
    )


def test_merge_into_update_multiple_columns():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN MATCHED THEN UPDATE SET target.v = src.v, target.v1 = src.v1"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("v", "src"), ColumnQualifierTuple("v", "target")),
            (ColumnQualifierTuple("v1", "src"), ColumnQualifierTuple("v1", "target")),
        ],
    )


def test_merge_into_update_multiple_columns_with_constant():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN MATCHED THEN UPDATE SET target.v = src.v, target.v1 = 1"""
    assert_column_lineage_equal(
        sql, [(ColumnQualifierTuple("v", "src"), ColumnQualifierTuple("v", "target"))]
    )
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN MATCHED THEN UPDATE SET target.v1 = 1, target.v = src.v"""
    assert_column_lineage_equal(
        sql, [(ColumnQualifierTuple("v", "src"), ColumnQualifierTuple("v", "target"))]
    )


def test_merge_into_update_multiple_match():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN MATCHED AND src.v0=1 THEN UPDATE SET target.v = src.v
WHEN MATCHED AND src.v0=2 THEN UPDATE SET target.v1 = src.v1"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("v", "src"), ColumnQualifierTuple("v", "target")),
            (ColumnQualifierTuple("v1", "src"), ColumnQualifierTuple("v1", "target")),
        ],
    )


def test_merge_into_insert():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN NOT MATCHED THEN INSERT (k, v) VALUES (src.k, src.v)"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("k", "src"), ColumnQualifierTuple("k", "target")),
            (ColumnQualifierTuple("v", "src"), ColumnQualifierTuple("v", "target")),
        ],
    )


def test_merge_into_insert_with_constant():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN NOT MATCHED THEN INSERT (k, v) VALUES (src.k, 1)"""
    assert_column_lineage_equal(
        sql, [(ColumnQualifierTuple("k", "src"), ColumnQualifierTuple("k", "target"))]
    )
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN NOT MATCHED THEN INSERT (v, k) VALUES (1, src.k)"""
    assert_column_lineage_equal(
        sql, [(ColumnQualifierTuple("k", "src"), ColumnQualifierTuple("k", "target"))]
    )


def test_merge_into_insert_one_column():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN NOT MATCHED THEN INSERT (k) VALUES (src.k)"""
    assert_column_lineage_equal(
        sql,
        [(ColumnQualifierTuple("k", "src"), ColumnQualifierTuple("k", "target"))],
    )


def test_merge_into_using_subquery():
    sql = """MERGE INTO target USING (select k, max(v) as v_max from src group by k) AS b ON target.k = b.k
WHEN MATCHED THEN UPDATE SET target.v = b.v_max
WHEN NOT MATCHED THEN INSERT (k, v) VALUES (b.k, b.v_max)"""
    assert_column_lineage_equal(
        sql,
        [
            (ColumnQualifierTuple("v", "src"), ColumnQualifierTuple("v", "target")),
            (ColumnQualifierTuple("k", "src"), ColumnQualifierTuple("k", "target")),
        ],
    )



import pytest

from ...helpers import assert_table_lineage_equal


@pytest.mark.parametrize("dialect", ["hive"])
def test_alter_table_exchange_partition(dialect: str):
    """
    See https://cwiki.apache.org/confluence/display/Hive/Exchange+Partition for language manual
    """
    assert_table_lineage_equal(
        "ALTER TABLE tab1 EXCHANGE PARTITION (pt='part1') WITH TABLE tab2",
        {"tab2"},
        {"tab1"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["snowflake"])
def test_alter_table_swap_partition(dialect: str):
    """
    See https://docs.snowflake.com/en/sql-reference/sql/alter-table for language manual
    Note swap is not a keyword in sqlparse, we'll skip testing for it.
    """
    assert_table_lineage_equal(
        "ALTER TABLE tab1 SWAP WITH tab2",
        {"tab2"},
        {"tab1"},
        dialect=dialect,
        test_sqlparse=False,
    )



import logging
import warnings
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple

from sqllineage import DEFAULT_DIALECT, SQLPARSE_DIALECT
from sqllineage.config import SQLLineageConfig
from sqllineage.core.holders import SQLLineageHolder
from sqllineage.core.metadata.dummy import DummyMetaDataProvider
from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.core.models import Column, Table
from sqllineage.core.parser.sqlfluff.analyzer import SqlFluffLineageAnalyzer
from sqllineage.core.parser.sqlparse.analyzer import SqlParseLineageAnalyzer
from sqllineage.drawing import draw_lineage_graph
from sqllineage.io import to_cytoscape
from sqllineage.utils.constant import LineageLevel
from sqllineage.utils.helpers import split, trim_comment

logger = logging.getLogger(__name__)


def lazy_method(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        if not self._evaluated:
            self._eval()
        return func(*args, **kwargs)

    return wrapper


def lazy_property(func):
    return property(lazy_method(func))


class LineageRunner(object):
    def __init__(
        self,
        sql: str,
        dialect: str = DEFAULT_DIALECT,
        metadata_provider: MetaDataProvider = DummyMetaDataProvider(),
        verbose: bool = False,
        silent_mode: bool = False,
        draw_options: Optional[Dict[str, str]] = None,
    ):
        """
        The entry point of SQLLineage after command line options are parsed.

        :param sql: a string representation of SQL statements.
        :param dialect: sql dialect
        :param metadata_provider: metadata service object providing table schema
        :param verbose: verbose flag indicating whether statement-wise lineage result will be shown
        :param silent_mode: boolean flag indicating whether to skip lineage analysis for unknown statement types
        """
        if dialect == SQLPARSE_DIALECT:
            warnings.warn(
                f"dialect `{SQLPARSE_DIALECT}` is deprecated, use `ansi` or dialect of your SQL instead. "
                f"`{SQLPARSE_DIALECT}` will be completely removed in v1.6.x",
                DeprecationWarning,
                stacklevel=2,
            )
        self._sql = sql
        self._verbose = verbose
        self._draw_options = draw_options if draw_options else {}
        self._evaluated = False
        self._stmt: List[str] = []
        self._dialect = dialect
        self._metadata_provider = metadata_provider
        self._silent_mode = silent_mode

    @lazy_method
    def __str__(self):
        """
        print out the Lineage Summary.
        """
        statements = self.statements()
        source_tables = "\n    ".join(str(t) for t in self.source_tables)
        target_tables = "\n    ".join(str(t) for t in self.target_tables)
        combined = f"""Statements(#): {len(statements)}
Source Tables:
    {source_tables}
Target Tables:
    {target_tables}
"""
        if self.intermediate_tables:
            intermediate_tables = "\n    ".join(
                str(t) for t in self.intermediate_tables
            )
            combined += f"""Intermediate Tables:
    {intermediate_tables}"""
        if self._verbose:
            result = ""
            for i, holder in enumerate(self._stmt_holders):
                stmt_short = statements[i].replace("\n", "")
                if len(stmt_short) > 50:
                    stmt_short = stmt_short[:50] + "..."
                content = str(holder).replace("\n", "\n    ")
                result += f"""Statement #{i + 1}: {stmt_short}
    {content}
"""
            combined = result + "==========\nSummary:\n" + combined
        return combined

    @lazy_method
    def to_cytoscape(self, level=LineageLevel.TABLE) -> List[Dict[str, Dict[str, str]]]:
        """
        to turn the DAG into cytoscape format.
        """
        if level == LineageLevel.COLUMN:
            return to_cytoscape(self._sql_holder.column_lineage_graph, compound=True)
        else:
            return to_cytoscape(self._sql_holder.table_lineage_graph)

    def draw(self) -> None:
        """
        to draw the lineage directed graph
        """
        draw_options = self._draw_options
        if draw_options.get("f") is None:
            draw_options.pop("f", None)
            draw_options["e"] = self._sql
            draw_options["dialect"] = self._dialect
        return draw_lineage_graph(**draw_options)

    @lazy_method
    def statements(self) -> List[str]:
        """
        a list of SQL statements.
        """
        return [trim_comment(s) for s in self._stmt]

    @lazy_property
    def source_tables(self) -> List[Table]:
        """
        a list of source :class:`sqllineage.models.Table`
        """
        return sorted(self._sql_holder.source_tables, key=lambda x: str(x))

    @lazy_property
    def target_tables(self) -> List[Table]:
        """
        a list of target :class:`sqllineage.models.Table`
        """
        return sorted(self._sql_holder.target_tables, key=lambda x: str(x))

    @lazy_property
    def intermediate_tables(self) -> List[Table]:
        """
        a list of intermediate :class:`sqllineage.models.Table`
        """
        return sorted(self._sql_holder.intermediate_tables, key=lambda x: str(x))

    @lazy_method
    def get_column_lineage(
        self, exclude_path_ending_in_subquery=True, exclude_subquery_columns=False
    ) -> List[Tuple[Column, Column]]:
        """
        a list of column tuple :class:`sqllineage.models.Column`
        """
        # sort by target column, and then source column
        return sorted(
            self._sql_holder.get_column_lineage(
                exclude_path_ending_in_subquery, exclude_subquery_columns
            ),
            key=lambda x: (str(x[-1]), str(x[0])),
        )

    def print_column_lineage(self) -> None:
        """
        print column level lineage to stdout
        """
        for path in self.get_column_lineage():
            print(" <- ".join(str(col) for col in reversed(path)))

    def print_table_lineage(self) -> None:
        """
        print table level lineage to stdout
        """
        print(str(self))

    def _eval(self):
        analyzer = (
            SqlParseLineageAnalyzer()
            if self._dialect == SQLPARSE_DIALECT
            else SqlFluffLineageAnalyzer(self._dialect, self._silent_mode)
        )
        if SQLLineageConfig.TSQL_NO_SEMICOLON and self._dialect == "tsql":
            self._stmt = analyzer.split_tsql(self._sql.strip())
        else:
            if SQLLineageConfig.TSQL_NO_SEMICOLON and self._dialect != "tsql":
                warnings.warn(
                    f"Dialect={self._dialect}, TSQL_NO_SEMICOLON will be ignored unless dialect is tsql"
                )
            self._stmt = split(self._sql.strip())

        with self._metadata_provider.session() as session:
            stmt_holders = []
            for stmt in self._stmt:
                stmt_holder = analyzer.analyze(stmt, session.metadata_provider)
                if write := stmt_holder.write:
                    tgt_table = next(iter(write))
                    if isinstance(tgt_table, Table) and (
                        tgt_columns := stmt_holder.get_table_columns(tgt_table)
                    ):
                        session.register_session_metadata(tgt_table, tgt_columns)
                stmt_holders.append(stmt_holder)
            self._stmt_holders = stmt_holders
            self._sql_holder = SQLLineageHolder.of(
                session.metadata_provider, *self._stmt_holders
            )
        self._evaluated = True

    @staticmethod
    def supported_dialects() -> Dict[str, List[str]]:
        """
        an ordered dict (so we can make sure the default parser implementation comes first)
        with key, value as parser_name, dialect list respectively
        """
        dialects = OrderedDict(
            [
                (
                    SqlFluffLineageAnalyzer.PARSER_NAME,
                    SqlFluffLineageAnalyzer.SUPPORTED_DIALECTS,
                ),
                (
                    SqlParseLineageAnalyzer.PARSER_NAME,
                    SqlParseLineageAnalyzer.SUPPORTED_DIALECTS,
                ),
            ]
        )
        return dialects



class SQLLineageException(Exception):
    """Base Exception for SQLLineage"""


class UnsupportedStatementException(SQLLineageException):
    """Raised for SQL statement that SQLLineage doesn't support analyzing"""


class InvalidSyntaxException(SQLLineageException):
    """Raised for SQL statement that parser cannot parse"""


class MetaDataProviderException(SQLLineageException):
    """Raised for MetaDataProvider errors"""


class ConfigException(SQLLineageException):
    """Raised for configuration errors"""



from ...helpers import assert_table_lineage_equal


def test_with_select():
    assert_table_lineage_equal("WITH tab1 AS (SELECT 1) SELECT * FROM tab1")


def test_with_select_one():
    assert_table_lineage_equal(
        "WITH wtab1 AS (SELECT * FROM schema1.tab1) SELECT * FROM wtab1",
        {"schema1.tab1"},
    )


def test_with_select_many():
    assert_table_lineage_equal(
        """WITH
cte1 AS (SELECT a, b FROM tab1),
cte2 AS (SELECT c, d FROM tab2)
SELECT b, d FROM cte1 JOIN cte2
WHERE cte1.a = cte2.c""",
        {"tab1", "tab2"},
    )


def test_with_select_many_reference():
    assert_table_lineage_equal(
        """WITH
cte1 AS (SELECT a, b FROM tab1),
cte2 AS (SELECT a, count(*) AS cnt FROM cte1 GROUP BY a)
SELECT a, b, cnt FROM cte1 JOIN cte2
WHERE cte1.a = cte2.a""",
        {"tab1"},
    )


def test_with_select_using_alias():
    assert_table_lineage_equal(
        "WITH wtab1 AS (SELECT * FROM schema1.tab1) SELECT * FROM wtab1 wt",
        {"schema1.tab1"},
    )


def test_with_select_join_table_with_same_name():
    assert_table_lineage_equal(
        "WITH wtab1 AS (SELECT * FROM schema1.tab1) SELECT * FROM wtab1 CROSS JOIN db.wtab1",
        {"schema1.tab1", "db.wtab1"},
    )


def test_cte_insert():
    assert_table_lineage_equal(
        "WITH tab1 AS (SELECT * FROM tab2) INSERT INTO tab3 SELECT * FROM tab1",
        {"tab2"},
        {"tab3"},
    )


def test_cte_insert_cte_in_query():
    assert_table_lineage_equal(
        "INSERT INTO tab3 WITH tab1 AS (SELECT * FROM tab2) SELECT * FROM tab1",
        {"tab2"},
        {"tab3"},
    )


def test_cte_update_from():
    assert_table_lineage_equal(
        """WITH cte1 AS (SELECT col1, col2 FROM tab1)
UPDATE tab2
SET tab2.col2 = cte1.col2 FROM cte1
WHERE tab2.col1 = cte1.col1;""",
        {"tab1"},
        {"tab2"},
    )


def test_cte_and_union():
    sql = """WITH cte_1 AS (select col1 from tab1)
SELECT col2 from tab2
UNION
SELECT col3 from cte_1"""
    assert_table_lineage_equal(
        sql,
        {"tab1", "tab2"},
    )


def test_cte_and_union_but_not_selecting_from_cte():
    # issue #398
    sql = """WITH cte_1 AS (select col1 from tab1)
SELECT col2 from tab2
UNION
SELECT col3 from tab3"""
    assert_table_lineage_equal(
        sql,
        {"tab1", "tab2", "tab3"},
    )


def test_cte_within_subquery():
    sql = """SELECT sq.col1
FROM (WITH cte1 AS (SELECT col1 FROM tab1)
      SELECT col1
      FROM cte1
               INNER JOIN tab2 ON cte1.col1 = tab2.col1) AS sq"""
    assert_table_lineage_equal(
        sql,
        {"tab1", "tab2"},
    )


def test_cte_within_cte():
    sql = """WITH cte1 AS
         (WITH cte2 AS
                   (SELECT id
                    FROM tab1)
          SELECT id
          FROM cte2)
SELECT id
FROM cte1"""
    assert_table_lineage_equal(sql, {"tab1"})


def test_non_reserved_keyword_as_cte():
    assert_table_lineage_equal(
        "WITH summary AS (SELECT * FROM segment) INSERT INTO host SELECT * FROM summary",
        {"segment"},
        {"host"},
        test_sqlparse=False,
    )



"""
Utils class to deal with the sqlfluff segments manipulations

naming convention:
    is_x: BaseSegment -> bool
    find_x: BaseSegment -> Optional[BaseSegment]
    list_x: BaseSegment -> List[BaseSegment]
    extract_x: other pattern
first parameter of each function must be sqlfluff BaseSegment
"""

from typing import List, Optional, Tuple

from sqlfluff.core.parser import BaseSegment

from sqllineage.utils.entities import ColumnQualifierTuple, SubQueryTuple


def is_negligible(segment: BaseSegment) -> bool:
    return (
        segment.is_whitespace
        or segment.is_comment
        or bool(segment.is_meta)
        or (segment.type == "symbol" and segment.raw != "*")
    )


def is_set_expression(segment: BaseSegment) -> bool:
    return segment.type == "set_expression" or any(
        seg.type == "set_expression" for seg in segment.segments
    )


def is_subquery(segment: BaseSegment) -> bool:
    if segment.type == "from_expression_element" or segment.type == "bracketed":
        token = extract_innermost_bracketed(
            segment if segment.type == "bracketed" else segment.segments[0]
        )
        # check if innermost parenthesis contains SELECT
        if token.get_child(
            "select_statement", "set_expression", "with_compound_statement"
        ):
            return True
        elif expression := token.get_child("expression"):
            if expression.get_child("select_statement"):
                return True
    return False


def is_wildcard(segment: BaseSegment) -> bool:
    return segment.type == "wildcard_expression" or (
        segment.type == "symbol" and segment.raw == "*" and segment.get_type() == "star"
    )


def find_from_expression_element(segment: BaseSegment) -> Optional[BaseSegment]:
    """
    segment can be of type:
        from_clause as grandparent
        from_expression/join_clause as parent
    """
    try:
        from_expression_element = next(
            segment.recursive_crawl("from_expression_element")
        )
    except StopIteration:
        from_expression_element = None
    return from_expression_element


def find_table_identifier(segment: BaseSegment) -> Optional[BaseSegment]:
    """
    recursively find table identifier
    """
    table_identifier = None
    if segment.type in ["table_reference", "file_reference", "object_reference"]:
        return segment
    else:
        for sub_segment in segment.segments:
            if identifier := find_table_identifier(sub_segment):
                return identifier
    return table_identifier


def list_join_clause(segment: BaseSegment) -> List[BaseSegment]:
    """
    traverse from_clause, recursively goes into bracket by default
    """
    if segment.type in ["from_clause", "update_statement"]:
        return list(segment.recursive_crawl("join_clause"))
    return []


def list_expression_from_when_clause(
    when_clause: BaseSegment, sub_type: str
) -> List[BaseSegment]:
    """
    sub_type can be either WHEN or THEN, this is to extract subquery for case when (SELECT ...) then (SELECT ...)
    """
    segments = []
    start_flag = False
    from_keyword, end_keyword = sub_type, "THEN" if sub_type == "WHEN" else None
    for segment in when_clause.segments:
        if segment.type == "keyword" and segment.raw_upper == from_keyword:
            start_flag = True
            continue
        if (
            end_keyword is not None
            and segment.type == "keyword"
            and segment.raw_upper == end_keyword
        ):
            break
        if start_flag and segment.type == "expression":
            segments = segment.get_children("bracketed")
    return segments


def list_subqueries(segment: BaseSegment) -> List[SubQueryTuple]:
    subquery = []
    if segment.type == "select_clause":
        for select_clause_element in segment.get_children("select_clause_element"):
            if expression := select_clause_element.get_child("expression"):
                if case_expression := expression.get_child("case_expression"):
                    for when_clause in case_expression.get_children("when_clause"):
                        for bracketed_segment in list_expression_from_when_clause(
                            when_clause, sub_type="WHEN"
                        ):
                            subquery.append(SubQueryTuple(bracketed_segment, None))
                        for bracketed_segment in list_expression_from_when_clause(
                            when_clause, sub_type="THEN"
                        ):
                            alias_expression = select_clause_element.get_child(
                                "alias_expression"
                            )
                            alias = (
                                extract_identifier(alias_expression)
                                if alias_expression
                                else None
                            )
                            subquery.append(SubQueryTuple(bracketed_segment, alias))
    elif segment.type == "from_expression_element":
        as_segment, target = extract_as_and_target_segment(segment)
        if is_subquery(target):
            subquery = [
                SubQueryTuple(
                    (
                        extract_innermost_bracketed(target)
                        if not is_set_expression(target)
                        else target
                    ),
                    extract_identifier(as_segment) if as_segment else None,
                )
            ]
    elif segment.type == "where_clause":
        bracketeds = []
        if expression := segment.get_child("expression"):
            bracketeds = expression.get_children("bracketed")
        elif bracketed_where := segment.get_child("bracketed"):
            if expression := bracketed_where.get_child("expression"):
                bracketeds = expression.get_children("bracketed")
        subquery = [
            SubQueryTuple(extract_innermost_bracketed(bracketed), None)
            for bracketed in bracketeds
            if is_subquery(bracketed)
        ]
    elif segment.type in ["from_clause", "from_expression"]:
        if from_expression_element := find_from_expression_element(segment):
            subquery = list_subqueries(from_expression_element)
        for join_clause in list_join_clause(segment):
            if from_expression_element := find_from_expression_element(join_clause):
                subquery += list_subqueries(from_expression_element)
    elif is_set_expression(segment):
        subquery = [
            SubQueryTuple(s, None)
            for s in list_child_segments(segment, check_bracketed=True)
            if s.type in ("bracketed", "select_statement")
        ]
    return subquery


def list_child_segments(
    segment: BaseSegment, check_bracketed: bool = True
) -> List[BaseSegment]:
    """
    Filter segments for a given segment's children, recursive goes into bracket by default
    """
    if segment.type == "bracketed" and check_bracketed:
        if is_set_expression(segment):
            return [seg for seg in segment.segments if seg.type == "set_expression"]
        else:
            result = []
            for seg in segment.iter_segments(
                expanding=["expression"], pass_through=True
            ):
                if seg.type in ["column_reference", "column_definition"]:
                    result.append(seg)
                else:
                    for s in seg.segments:
                        if not is_negligible(s):
                            result.append(s)
            return result
    else:
        return [seg for seg in segment.segments if not is_negligible(seg)]


def extract_identifier(col_segment: BaseSegment) -> str:
    identifiers = list_child_segments(col_segment)
    col_identifier = identifiers[-1]
    return str(col_identifier.raw)


def extract_as_and_target_segment(
    segment: BaseSegment,
) -> Tuple[Optional[BaseSegment], BaseSegment]:
    as_segment = segment.get_child("alias_expression")
    sublist = list_child_segments(segment, False)
    target = sublist[0] if is_subquery(sublist[0]) else sublist[0].segments[0]
    return as_segment, target


def extract_column_qualifier(segment: BaseSegment) -> Optional[ColumnQualifierTuple]:
    cqt = None
    if is_wildcard(segment):
        identifiers = segment.raw.split(".")
        column = identifiers[-1]
        parent = identifiers[-2] if len(identifiers) > 1 else None
        cqt = ColumnQualifierTuple(column, parent)
    elif segment.type == "column_reference":
        sub_segments = list_child_segments(segment)
        column = sub_segments[-1].raw
        parent = sub_segments[-2].raw if len(sub_segments) > 1 else None
        cqt = ColumnQualifierTuple(column, parent)
    elif segment.type == "identifier":
        cqt = ColumnQualifierTuple(segment.raw, None)
    return cqt


def extract_innermost_bracketed(bracketed_segment: BaseSegment) -> BaseSegment:
    # in case of subquery in nested parenthesis like: `SELECT * FROM ((table))`, find the innermost one first
    while True:
        sub_bracketed_segments = [
            bs.get_child("bracketed")
            for bs in bracketed_segment.segments
            if bs.get_child("bracketed")
        ]
        sub_paren = bracketed_segment.get_child("bracketed") or (
            sub_bracketed_segments[0] if sub_bracketed_segments else None
        )
        if sub_paren is not None:
            bracketed_segment = sub_paren
        else:
            break
    return bracketed_segment



from ...helpers import assert_table_lineage_equal


def test_merge_into_using_table():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN MATCHED THEN UPDATE SET target.v = src.v"""
    assert_table_lineage_equal(sql, {"src"}, {"target"})


def test_merge_into_using_subquery():
    sql = """MERGE INTO target USING (select k, max(v) as v from src group by k) AS b ON target.k = b.k
WHEN MATCHED THEN UPDATE SET target.v = b.v
WHEN NOT MATCHED THEN INSERT (k, v) VALUES (b.k, b.v)"""
    assert_table_lineage_equal(sql, {"src"}, {"target"})


def test_merge_using_cte_subquery():
    sql = """MERGE INTO tgt t
USING (
    WITH base AS (
        SELECT
            id, max(value) AS value
        FROM src
        GROUP BY id
    )
    SELECT
        id, value
    FROM base
) s
ON t.id = s.id
WHEN MATCHED THEN
UPDATE SET t.value = s.value"""
    assert_table_lineage_equal(
        sql,
        {"src"},
        {"tgt"},
    )


def test_merge_into_insert_one_column():
    sql = """MERGE INTO target
USING src ON target.k = src.k
WHEN NOT MATCHED THEN INSERT VALUES (src.k)"""
    assert_table_lineage_equal(sql, {"src"}, {"target"})


def test_merge_with_union_in_subquery_and_join():
    sql = """MERGE INTO tgt t
USING (SELECT s1.id, baz.value
       FROM (SELECT id
             FROM foo
             UNION ALL
             SELECT id
             FROM bar) s1
                CROSS JOIN baz) s
ON t.id = s.id
WHEN MATCHED THEN UPDATE SET t.value = s.value
"""
    assert_table_lineage_equal(sql, {"foo", "bar", "baz"}, {"tgt"})



from sqlfluff.core.parser import BaseSegment

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.core.parser import SourceHandlerMixin
from sqllineage.core.parser.sqlfluff.extractors.base import BaseExtractor
from sqllineage.core.parser.sqlfluff.models import (
    SqlFluffColumn,
    SqlFluffTable,
)
from sqllineage.core.parser.sqlfluff.utils import (
    find_table_identifier,
    is_set_expression,
    list_child_segments,
)
from sqllineage.utils.entities import AnalyzerContext
from sqllineage.utils.helpers import escape_identifier_name


class SelectExtractor(BaseExtractor, SourceHandlerMixin):
    """
    Select statement lineage extractor
    """

    SUPPORTED_STMT_TYPES = ["select_statement", "set_expression", "bracketed"]

    def __init__(self, dialect: str, metadata_provider: MetaDataProvider):
        super().__init__(dialect, metadata_provider)
        self.columns = []
        self.tables = []
        self.union_barriers = []

    def extract(
        self,
        statement: BaseSegment,
        context: AnalyzerContext,
    ) -> SubQueryLineageHolder:
        holder = self._init_holder(context)
        subqueries = []
        segments = (
            [statement]
            if statement.type == "set_expression"
            else list_child_segments(statement)
        )
        for segment in segments:
            for sq in self.list_subquery(segment):
                # Collecting subquery on the way, hold on parsing until last
                # so that each handler don't have to worry about what's inside subquery
                subqueries.append(sq)

            if is_set_expression(segment):
                for _, sub_segment in enumerate(
                    segment.get_children("select_statement", "bracketed")
                ):
                    for seg in list_child_segments(sub_segment):
                        for sq in self.list_subquery(seg):
                            subqueries.append(sq)

        self.extract_subquery(subqueries, holder)

        for segment in segments:
            self._handle_select_statement_child_segments(segment, holder)

            if is_set_expression(segment):
                for idx, sub_segment in enumerate(
                    segment.get_children("select_statement", "bracketed")
                ):
                    if idx != 0:
                        self.union_barriers.append(
                            (len(self.columns), len(self.tables))
                        )
                    for seg in list_child_segments(sub_segment):
                        self._handle_select_statement_child_segments(seg, holder)

        self.end_of_query_cleanup(holder)

        holder.expand_wildcard(self.metadata_provider)

        return holder

    def _handle_select_statement_child_segments(
        self, segment: BaseSegment, holder: SubQueryLineageHolder
    ):
        self._handle_swap_partition(segment, holder)
        self._handle_select_into(segment, holder)
        self.tables.extend(
            self._list_table_from_from_clause_or_join_clause(segment, holder)
        )
        self._handle_column(segment)

    @staticmethod
    def _handle_swap_partition(segment: BaseSegment, holder: SubQueryLineageHolder):
        """
        A handler for swap_partitions_between_tables function
        """
        if segment.type == "select_clause":
            if select_clause_element := segment.get_child("select_clause_element"):
                if function := select_clause_element.get_child("function"):
                    if (
                        function.first_non_whitespace_segment_raw_upper
                        == "SWAP_PARTITIONS_BETWEEN_TABLES"
                    ):
                        if bracketed := function.get_child("bracketed"):
                            expressions = bracketed.get_children("expression")
                            holder.add_read(
                                SqlFluffTable(
                                    escape_identifier_name(expressions[0].raw)
                                )
                            )
                            holder.add_write(
                                SqlFluffTable(
                                    escape_identifier_name(expressions[3].raw)
                                )
                            )

    def _handle_select_into(self, segment: BaseSegment, holder: SubQueryLineageHolder):
        """
        A handler for SELECT INTO statement supported by some dialects
        """
        if segment.type in ["into_table_clause", "into_clause"]:
            if identifier := find_table_identifier(segment):
                if table := self.find_table(identifier):
                    holder.add_write(table)

    def _handle_column(self, segment: BaseSegment) -> None:
        """
        Column handler method
        """
        if segment.type == "select_clause":
            for sub_segment in segment.get_children("select_clause_element"):
                self.columns.append(SqlFluffColumn.of(sub_segment))



from functools import reduce
from operator import add
from typing import List, Optional, Union

import sqlparse
from sqlparse.sql import (
    Comparison,
    Function,
    Identifier,
    IdentifierList,
    Parenthesis,
    Statement,
    TokenList,
    Values,
    Where,
)

from sqllineage import SQLPARSE_DIALECT
from sqllineage.core.analyzer import LineageAnalyzer
from sqllineage.core.holders import StatementLineageHolder, SubQueryLineageHolder
from sqllineage.core.metadata_provider import MetaDataProvider
from sqllineage.core.models import Column, SubQuery, Table
from sqllineage.core.parser.sqlparse.handlers.base import (
    CurrentTokenBaseHandler,
    NextTokenBaseHandler,
)
from sqllineage.core.parser.sqlparse.models import SqlParseSubQuery, SqlParseTable
from sqllineage.core.parser.sqlparse.utils import (
    get_subquery_parentheses,
    is_subquery,
    is_token_negligible,
)
from sqllineage.utils.entities import AnalyzerContext
from sqllineage.utils.helpers import trim_comment


class SqlParseLineageAnalyzer(LineageAnalyzer):
    """SQL Statement Level Lineage Analyzer."""

    PARSER_NAME = "sqlparse"
    SUPPORTED_DIALECTS = [SQLPARSE_DIALECT]

    def analyze(
        self, sql: str, metadata_provider: MetaDataProvider
    ) -> StatementLineageHolder:
        # get rid of comments, which cause inconsistencies in sqlparse output
        stmt = sqlparse.parse(trim_comment(sql))[0]
        if (
            stmt.get_type() == "DELETE"
            or stmt.token_first(skip_cm=True).normalized == "TRUNCATE"
            or stmt.token_first(skip_cm=True).normalized.upper() == "REFRESH"
            or stmt.token_first(skip_cm=True).normalized == "CACHE"
            or stmt.token_first(skip_cm=True).normalized.upper() == "UNCACHE"
            or stmt.token_first(skip_cm=True).normalized == "SHOW"
        ):
            holder = StatementLineageHolder()
        elif stmt.get_type() == "DROP":
            holder = self._extract_from_ddl_drop(stmt)
        elif (
            stmt.get_type() == "ALTER"
            or stmt.token_first(skip_cm=True).normalized == "RENAME"
        ):
            holder = self._extract_from_ddl_alter(stmt)
        elif stmt.get_type() == "MERGE":
            holder = self._extract_from_dml_merge(stmt, metadata_provider)
        else:
            # DML parsing logic also applies to CREATE DDL
            holder = StatementLineageHolder.of(
                self._extract_from_dml(stmt, AnalyzerContext(), metadata_provider)
            )
        return holder

    @classmethod
    def _extract_from_ddl_drop(cls, stmt: Statement) -> StatementLineageHolder:
        holder = StatementLineageHolder()
        for table in {
            SqlParseTable.of(t) for t in stmt.tokens if isinstance(t, Identifier)
        }:
            holder.add_drop(table)
        return holder

    @classmethod
    def _extract_from_ddl_alter(cls, stmt: Statement) -> StatementLineageHolder:
        holder = StatementLineageHolder()
        tables = []
        for t in stmt.tokens:
            if isinstance(t, Identifier):
                tables.append(SqlParseTable.of(t))
            elif isinstance(t, IdentifierList):
                for identifier in t.get_identifiers():
                    tables.append(SqlParseTable.of(identifier))
        keywords = [t for t in stmt.tokens if t.is_keyword]
        if any(k.normalized == "RENAME" for k in keywords):
            if stmt.get_type() == "ALTER" and len(tables) == 2:
                holder.add_rename(tables[0], tables[1])
            elif (
                stmt.token_first(skip_cm=True).normalized == "RENAME"
                and len(tables) % 2 == 0
            ):
                for i in range(0, len(tables), 2):
                    holder.add_rename(tables[i], tables[i + 1])
        if any(k.normalized == "EXCHANGE" for k in keywords) and len(tables) == 2:
            holder.add_write(tables[0])
            holder.add_read(tables[1])
        return holder

    @classmethod
    def _extract_from_dml_merge(
        cls,
        stmt: Statement,
        metadata_provider: MetaDataProvider,
    ) -> StatementLineageHolder:
        holder = StatementLineageHolder()
        src_flag = tgt_flag = update_flag = insert_flag = False
        insert_columns = []
        direct_source: Optional[Union[Table, SubQuery]] = None
        for token in stmt.tokens:
            if is_token_negligible(token):
                continue
            if token.is_keyword:
                if token.normalized in {"INTO", "MERGE"}:
                    tgt_flag = True
                elif token.normalized == "USING":
                    src_flag = True
                elif token.normalized == "SET":
                    update_flag = True
                elif token.normalized == "INSERT":
                    insert_flag = True
                elif token.normalized == "WHEN":
                    update_flag = False
                continue

            if tgt_flag:
                if isinstance(token, Identifier):
                    holder.add_write(SqlParseTable.of(token))
                tgt_flag = False
            elif src_flag:
                if isinstance(token, Identifier):
                    if subqueries := cls.parse_subquery(token):
                        for sq in subqueries:
                            direct_source = SqlParseSubQuery.of(sq.query, sq.alias)
                            holder |= cls._extract_from_dml(
                                sq.query,
                                AnalyzerContext(cte=holder.cte, write={sq}),
                                metadata_provider,
                            )
                    else:
                        direct_source = SqlParseTable.of(token)
                    holder.add_read(direct_source)
                src_flag = False
            elif update_flag:
                comparisons = []
                if isinstance(token, Comparison):
                    comparisons = [token]
                elif isinstance(token, IdentifierList):
                    comparisons = [
                        t for t in token.get_identifiers() if isinstance(t, Comparison)
                    ]
                for c in comparisons:
                    if isinstance(right := c.right, Identifier):
                        src_col = Column(right.get_real_name())
                        src_col.parent = direct_source
                        tgt_col = Column(c.left.get_real_name())
                        tgt_col.parent = list(holder.write)[0]
                        holder.add_column_lineage(src_col, tgt_col)
            elif insert_flag:
                if isinstance(token, Parenthesis):
                    t = token.tokens[1]
                    identifiers = []
                    if isinstance(t, Identifier):
                        identifiers.append(t)
                    elif isinstance(t, IdentifierList):
                        identifiers.extend(t.get_identifiers())
                    for identifier in identifiers:
                        tgt_col = Column(identifier.get_real_name())
                        tgt_col.parent = list(holder.write)[0]
                        insert_columns.append(tgt_col)
                elif insert_columns and isinstance(token, Values):
                    for sub_token in token.tokens:
                        if isinstance(sub_token, Parenthesis):
                            t = sub_token.tokens[1]
                            identifiers = []
                            if isinstance(t, Identifier):
                                identifiers.append(t)
                            elif isinstance(t, IdentifierList):
                                identifiers.extend(t.get_identifiers())
                            for i, identifier in enumerate(identifiers):
                                if isinstance(identifier, Identifier):
                                    src_col = Column(identifier.get_real_name())
                                    src_col.parent = direct_source
                                    holder.add_column_lineage(
                                        src_col, insert_columns[i]
                                    )
                    insert_flag = False
        return holder

    @classmethod
    def _extract_from_dml(
        cls,
        token: TokenList,
        context: AnalyzerContext,
        metadata_provider: MetaDataProvider,
    ) -> SubQueryLineageHolder:
        holder = SubQueryLineageHolder()
        if context.cte is not None:
            # CTE can be referenced by subsequent CTEs
            for cte in context.cte:
                holder.add_cte(cte)
        if context.write is not None:
            # If within subquery, then manually add subquery as target table
            for write in context.write:
                holder.add_write(write)
        current_handlers = [
            handler_cls() for handler_cls in CurrentTokenBaseHandler.__subclasses__()
        ]
        next_handlers = [
            handler_cls() for handler_cls in NextTokenBaseHandler.__subclasses__()
        ]

        subqueries = []
        for sub_token in token.tokens:
            if is_token_negligible(sub_token):
                continue

            for sq in cls.parse_subquery(sub_token):
                # Collecting subquery on the way, hold on parsing until last
                # so that each handler don't have to worry about what's inside subquery
                subqueries.append(sq)

            for current_handler in current_handlers:
                current_handler.handle(sub_token, holder)

            if sub_token.is_keyword:
                for next_handler in next_handlers:
                    next_handler.indicate(sub_token)
                continue

            for next_handler in next_handlers:
                if next_handler.indicator:
                    next_handler.handle(sub_token, holder)
        else:
            # call end of query hook here as loop is over
            for next_handler in next_handlers:
                next_handler.end_of_query_cleanup(holder)
        # By recursively extracting each subquery of the parent and merge, we're doing Depth-first search
        for sq in subqueries:
            holder |= cls._extract_from_dml(
                sq.query,
                AnalyzerContext(cte=holder.cte, write={sq}),
                metadata_provider,
            )
        # replace wildcard with real columns, put here so that wildcard in subqueries are already replaced
        holder.expand_wildcard(metadata_provider)
        return holder

    @classmethod
    def parse_subquery(cls, token: TokenList) -> List[SubQuery]:
        result = []
        if isinstance(token, (Identifier, Function, Where, Values)):
            # usually SubQuery is an Identifier, but not all Identifiers are SubQuery
            # Function for CTE without AS keyword
            result = cls._parse_subquery(token)
        elif isinstance(token, IdentifierList):
            # IdentifierList for SQL89 style of JOIN or multiple CTEs, this is actually SubQueries
            result = reduce(
                add,
                [
                    cls._parse_subquery(identifier)
                    for identifier in token.get_sublists()
                ],
                [],
            )
        elif is_subquery(token):
            # Parenthesis for SubQuery without alias, this is valid syntax for certain SQL dialect
            result = [SqlParseSubQuery.of(token, None)]
        return result

    @classmethod
    def _parse_subquery(
        cls, token: Union[Identifier, Function, Where]
    ) -> List[SubQuery]:
        """
        convert SubQueryTuple to sqllineage.core.models.SubQuery
        """
        return [
            SqlParseSubQuery.of(parenthesis, alias)
            for parenthesis, alias in get_subquery_parentheses(token)
        ]



import pytest

from ...helpers import assert_table_lineage_equal


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_create_bucket_table_in_spark(dialect: str):
    assert_table_lineage_equal(
        """CREATE TABLE student (id INT, name STRING, age INT)
USING CSV
PARTITIONED BY (age)
CLUSTERED BY (Id) INTO 4 buckets""",
        None,
        {"student"},
        dialect,
    )


@pytest.mark.parametrize("dialect", ["athena"])
def test_create_bucket_table_in_athena(dialect: str):
    assert_table_lineage_equal(
        """CREATE TABLE bar
WITH (
  bucketed_by = ARRAY['customer_id'],
  bucket_count = 8
)
AS SELECT * FROM foo""",
        {"foo"},
        {"bar"},
        dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "sparksql"])
def test_create_select_without_as(dialect: str):
    assert_table_lineage_equal(
        "CREATE TABLE tab1 SELECT * FROM tab2", {"tab2"}, {"tab1"}, dialect
    )


@pytest.mark.parametrize("dialect", ["duckdb", "greenplum", "postgres", "redshift"])
def test_create_table_as_with_postgres_dialect(dialect: str):
    """
    sqlfluff postgres family dialects parse CTAS statement as "create_table_as_statement",
    unlike "create_table_statement" in ansi dialect
    """
    assert_table_lineage_equal(
        """CREATE TABLE bar AS
SELECT *
FROM foo""",
        {"foo"},
        {"bar"},
        dialect,
    )


@pytest.mark.parametrize("dialect", ["databricks", "hive", "sparksql"])
def test_create_using_serde(dialect: str):
    """
    https://cwiki.apache.org/confluence/display/Hive/LanguageManual+DDL#LanguageManualDDL-RowFormats&SerDe
    here with is not an indicator for CTE
    """
    # use raw sql string to avoid Python recognize some character as escaping
    assert_table_lineage_equal(
        r"""CREATE TABLE apachelog (
  host STRING,
  identity STRING,
  `user` STRING,
  `time` STRING,
  request STRING,
  status STRING,
  size STRING,
  referer STRING,
  agent STRING)
ROW FORMAT SERDE 'org.apache.hadoop.hive.serde2.RegexSerDe'
WITH SERDEPROPERTIES (
  "input.regex" = "([^]*) ([^]*) ([^]*) (-|\\[^\\]*\\]) ([^ \"]*|\"[^\"]*\") (-|[0-9]*) (-|[0-9]*)(?: ([^ \"]*|\".*\") ([^ \"]*|\".*\"))?"
)
STORED AS TEXTFILE""",  # noqa
        None,
        {"apachelog"},
        dialect=dialect,
    )


@pytest.mark.parametrize("dialect", ["snowflake", "bigquery"])
def test_create_clone(dialect: str):
    """
    Language manual:
        https://cloud.google.com/bigquery/docs/table-clones-create
        https://docs.snowflake.com/en/sql-reference/sql/create-clone
    Note clone is not a keyword in sqlparse, we'll skip testing for it.
    """
    assert_table_lineage_equal(
        "CREATE TABLE tab2 CLONE tab1",
        {"tab1"},
        {"tab2"},
        dialect=dialect,
        test_sqlparse=False,
    )



from sqllineage.utils.entities import ColumnQualifierTuple
from ...helpers import assert_column_lineage_equal


def test_column_reference_using_union():
    sql = """INSERT INTO tab3
SELECT col1
FROM tab1
UNION ALL
SELECT col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab1"),
                ColumnQualifierTuple("col1", "tab3"),
            ),
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab3"),
            ),
        ],
    )
    sql = """INSERT INTO tab3
SELECT col1
FROM tab1
UNION
SELECT col1
FROM tab2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "tab1"),
                ColumnQualifierTuple("col1", "tab3"),
            ),
            (
                ColumnQualifierTuple("col1", "tab2"),
                ColumnQualifierTuple("col1", "tab3"),
            ),
        ],
    )


def test_union_inside_cte():
    sql = """INSERT INTO dataset.target WITH temp_cte AS (SELECT col1 FROM dataset.tab1 UNION ALL
SELECT col1 FROM dataset.tab2) SELECT col1 FROM temp_cte"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("col1", "dataset.tab1"),
                ColumnQualifierTuple("col1", "dataset.target"),
            ),
            (
                ColumnQualifierTuple("col1", "dataset.tab2"),
                ColumnQualifierTuple("col1", "dataset.target"),
            ),
        ],
    )


def test_union_with_subquery():
    sql = """INSERT INTO tab3
SELECT sq1.id
FROM (SELECT id
      FROM tab1) sq1
UNION ALL
SELECT sq2.id
FROM (SELECT id
      FROM tab2) sq2"""
    assert_column_lineage_equal(
        sql,
        [
            (
                ColumnQualifierTuple("id", "tab1"),
                ColumnQualifierTuple("id", "tab3"),
            ),
            (
                ColumnQualifierTuple("id", "tab2"),
                ColumnQualifierTuple("id", "tab3"),
            ),
        ],
    )
